import shutil
import sys
import time
import unittest
from pathlib import Path

from CADETProcess import settings

sys.path.insert(0, "../")

# Set flags for which test cases to run
test_batch_elution_single_objective_single_core = True
test_batch_elution_single_objective_multi_core = True
test_batch_elution_multi_objective = True
test_fit_column_parameters = False


class TestBatchElutionOptimizationSingleObjective(unittest.TestCase):
    def setUp(self):
        self.tearDown()

        if not (
            test_batch_elution_single_objective_single_core
            or test_batch_elution_single_objective_multi_core
        ):
            return

        settings.working_directory = "./test_batch"
        settings.temp_dir = "./test_batch/tmp"

        from examples.batch_elution.optimization_single import (
            optimization_problem,
            optimizer,
        )

        optimization_problem.cache_directory = (
            "./test_batch/diskcache_batch_elution_single"
        )

        self.optimization_problem = optimization_problem
        self.optimizer = optimizer

    def tearDown(self):
        shutil.rmtree("./test_batch", ignore_errors=True)
        shutil.rmtree("./diskcache_batch_elution_single", ignore_errors=True)
        shutil.rmtree("./tmp", ignore_errors=True)

        settings.working_directory = None

    def test_single_core(self):
        if not test_batch_elution_single_objective_single_core:
            self.skipTest("Skipping test_batch_elution_single_objective_single_core")

        self.optimizer.n_cores = 1
        self.optimizer.pop_size = 4
        self.optimizer.n_max_gen = 4

        print("start test_batch_elution_single_objective_single_core")

        start = time.time()
        results = self.optimizer.optimize(
            self.optimization_problem,
            use_checkpoint=False,
        )
        end = time.time()

        print(
            f"Finalized test_batch_elution_single_objective_single_core in {end - start} s"
        )

    @unittest.skipIf(__name__ != "__main__", "Only run test if test is run as __main__")
    def test_multi_core(self):
        if not test_batch_elution_single_objective_multi_core:
            self.skipTest("Skipping test_batch_elution_single_objective_multi_core")

        self.optimizer.n_cores = -2
        self.optimizer.pop_size = 16
        self.optimizer.n_max_gen = 2

        print("start test_batch_elution_single_objective_multi_core")

        start = time.time()
        results = self.optimizer.optimize(
            self.optimization_problem,
            use_checkpoint=False,
        )
        end = time.time()

        print(
            f"Finalized test_batch_elution_single_objective_multi_core in {end - start} s"
        )
        print(f"Equivalent CPU time: {(end - start) * self.optimizer.n_cores} s")


class TestBatchElutionOptimizationMultiObjective(unittest.TestCase):
    def setUp(self):
        self.tearDown()

        if not test_batch_elution_multi_objective:
            return

        settings.working_directory = "./test_batch"
        settings.temp_dir = "./test_batch/tmp"

        from examples.batch_elution.optimization_multi import (
            optimization_problem,
            optimizer,
        )

        optimization_problem.cache_directory = (
            "./test_batch/diskcache_batch_elution_multi"
        )

        self.optimization_problem = optimization_problem
        self.optimizer = optimizer

    def tearDown(self):
        shutil.rmtree("./test_batch", ignore_errors=True)
        shutil.rmtree("./diskcache_batch_elution_multi", ignore_errors=True)
        shutil.rmtree("./tmp", ignore_errors=True)

        settings.working_directory = None

    @unittest.skipIf(__name__ != "__main__", "Only run test if test is run as __main__")
    def test_optimization(self):
        if not test_batch_elution_multi_objective:
            self.skipTest("Skipping test_batch_elution_multi_objective")

        self.optimizer.n_cores = -2
        self.optimizer.pop_size = 16
        self.optimizer.n_max_gen = 2

        print("start test_batch_elution_multi_objective")

        start = time.time()
        results = self.optimizer.optimize(
            self.optimization_problem,
            use_checkpoint=False,
        )
        end = time.time()

        print(f"Finalized test_batch_elution_multi_objective in {end - start} s")
        print(f"Equivalent CPU time: {(end - start) * self.optimizer.n_cores} s")

        print(results.f)
        print(results.x)


class TestFitColumnParameters(unittest.TestCase):
    def setUp(self):
        self.tearDown()

        if not test_fit_column_parameters:
            return

        settings.working_directory = "./test_fit_column_parameters"
        settings.temp_dir = "./test_fit_column_parameters/tmp"

        data_dir = (
            Path(__file__).parent.parent
            / "examples/characterize_chromatographic_system/experimental_data"
        )
        shutil.copytree(data_dir, "./experimental_data")

        from examples.characterize_chromatographic_system.column_transport_parameters import (
            optimization_problem,
            optimizer,
        )

        optimization_problem.cache_directory = (
            "./test_fit_column_parameters/diskcache_bed_porosity_axial_dispersion"
        )

        self.optimization_problem = optimization_problem
        self.optimizer = optimizer

    def tearDown(self):
        shutil.rmtree("./test_fit_column_parameters", ignore_errors=True)
        shutil.rmtree("./diskcache_bed_porosity_axial_dispersion/", ignore_errors=True)
        shutil.rmtree("./tmp", ignore_errors=True)

        shutil.rmtree("./experimental_data/", ignore_errors=True)

        settings.working_directory = None

    def test_optimization(self):
        if not test_fit_column_parameters:
            self.skipTest("Skipping test_fit_column_parameters")

        self.optimizer.n_cores = -2
        self.optimizer.pop_size = 16
        self.optimizer.n_max_gen = 2

        print("start test_fit_column_parameters")

        start = time.time()
        results = self.optimizer.optimize(
            self.optimization_problem,
            use_checkpoint=False,
        )
        end = time.time()

        print(f"Finalized test_fit_column_parameters in {end - start} s")
        print(f"Equivalent CPU time: {(end - start) * self.optimizer.n_cores} s")


if __name__ == "__main__":
    # Run the tests
    unittest.main()



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% tags=["remove-cell"]
from pathlib import Path
import sys

root_dir = Path('../../../../').resolve()
sys.path.append(root_dir.as_posix())

# %% [markdown]
# ## Fit Binding Model Parameters
#
# This example demonstrates how to estimate SMA binding parameters based on multiple gradient elution chromatograms.
# It assumes familiarity with the `CADETProcess` process models, as introduced in the
# **Fit Column Transport Parameters** example.

# %%
import os

import numpy as np

from CADETProcess.processModel import ComponentSystem, FlowSheet, Process, Inlet, Outlet, LumpedRateModelWithPores
from CADETProcess.processModel import StericMassAction
from CADETProcess.reference import ReferenceIO
from CADETProcess.comparison import Comparator
from CADETProcess.simulator import Cadet
from CADETProcess.optimization import OptimizationProblem


# %% [markdown]
# ## Process creation
# To simplify the creation of multiple `Process` instances with the correct configurations,
# the process creation was wrapped into a function below:

# %%
def create_process(cv_length=30):
    # Component System
    component_system = ComponentSystem()
    component_system.add_component('Salt')
    component_system.add_component('A')

    Q_ml_min = 0.5  # ml/min
    Q_m3_s = Q_ml_min / (60 * 1e6)

    initial_salt_concentration = 50
    final_salt_concentration = 500

    # Unit Operations
    inlet = Inlet(component_system, name='inlet')
    inlet.flow_rate = Q_m3_s

    column = create_column_model(component_system, final_salt_concentration, initial_salt_concentration)

    column_volume = column.length * (column.diameter / 2) ** 2 * np.pi

    outlet = Outlet(component_system, name='outlet')

    flow_sheet = FlowSheet(component_system)

    flow_sheet.add_unit(inlet)
    flow_sheet.add_unit(column)
    flow_sheet.add_unit(outlet)

    flow_sheet.add_connection(inlet, column)
    flow_sheet.add_connection(column, outlet)

    # Process
    process = Process(flow_sheet, f'{cv_length}')

    load_duration = 9
    t_gradient_start = 90.0
    gradient_volume = cv_length * column_volume
    gradient_duration = gradient_volume / inlet.flow_rate[0]
    duration_post_gradient_wash = gradient_duration / 10 + 180
    process.cycle_time = t_gradient_start + gradient_duration + duration_post_gradient_wash
    t_gradient_end = t_gradient_start + gradient_duration

    c_load = np.array([initial_salt_concentration, 1.0])
    c_wash = np.array([initial_salt_concentration, 0.0])
    c_elute = np.array([final_salt_concentration, 0.0])
    gradient_slope = (c_elute - c_wash) / gradient_duration
    c_gradient_poly = np.array(list(zip(c_wash, gradient_slope)))
    c_post_gradient_wash = np.array([final_salt_concentration, 0.0])

    process.add_event('load', 'flow_sheet.inlet.c', c_load)
    process.add_event('wash', 'flow_sheet.inlet.c', c_wash, load_duration)
    process.add_event('grad_start', 'flow_sheet.inlet.c', c_gradient_poly, t_gradient_start)
    process.add_event('grad_end', 'flow_sheet.inlet.c', c_post_gradient_wash, t_gradient_end)

    return process


def create_column_model(component_system, final_salt_concentration, initial_salt_concentration):
    column = LumpedRateModelWithPores(component_system, name='column')
    column.length = 0.02  # m
    column.diameter = 0.008
    column.particle_radius = 1.7e-5
    column.axial_dispersion = 3.16e-7
    column.bed_porosity = 0.3
    column.particle_porosity = 0.5
    column.film_diffusion = [1e-4, 1e-4]
    binding_model = create_binding_model(component_system, final_salt_concentration)
    column.binding_model = binding_model
    column.c = [initial_salt_concentration, 0, ]
    column.cp = [initial_salt_concentration, 0]
    column.q = [binding_model.capacity, 0]
    return column


def create_binding_model(component_system, final_salt_concentration):
    binding_model = StericMassAction(component_system, name='SMA')
    binding_model.is_kinetic = True
    binding_model.characteristic_charge = [0.0, 10]
    binding_model.capacity = 400.0
    binding_model.reference_solid_phase_conc = binding_model.capacity
    binding_model.reference_liquid_phase_conc = final_salt_concentration
    binding_model.steric_factor = [0.0, 10]
    binding_model.adsorption_rate = [0.0, 10]
    binding_model.desorption_rate = [0.0, 1e3]
    return binding_model


# %% [markdown]
# ## Creating "experimental" data
#
# To run the parameter estimation algorithm, we need experimental data. This function generates _in-silico_ based
# "experimental" data for us to use.

# %%
def create_in_silico_experimental_data():
    def save_csv(results, directory=None, filename=None, units=None, noise_percentage=5):
        if not os.path.exists(directory):
            os.makedirs(directory)

        if units is None:
            units = results.solution.keys()

        for unit in units:
            solution = results.solution[unit]["outlet"].solution
            solution_without_salt = solution[:, 1:]
            solution_without_salt = (solution_without_salt + (np.random.random(solution_without_salt.shape) - 0.5)
                                     * noise_percentage / 100 * solution_without_salt.max(axis=0)
                                     )
            solution[:, 1:] = solution_without_salt
            solution_times = results.solution[unit]["outlet"].time

            full_solution = np.concatenate([np.atleast_2d(solution_times), solution.T]).T

            if filename is None:
                filename = unit + '_output.csv'

            file_path = os.path.join(directory, filename)
            with open(file_path, "w") as file_handle:
                np.savetxt(file_handle, full_solution, delimiter=",")

    simulator = Cadet()
    case_dir = "experimental_data"
    os.makedirs(case_dir, exist_ok=True)

    for cv in [5, 30, 120]:
        process = create_process(cv_length=cv)
        simulation_results = simulator.simulate(process)
        save_csv(simulation_results, directory=case_dir, filename=f"{cv}.csv", units=["outlet"])


# %% [markdown]
# Below are two convenience functions, written to simplifly loading `references` and creating `comparators`.

# %%
def load_reference(file_name, component_index=2):
    data = np.loadtxt(file_name, delimiter=',')

    time_experiment = data[:, 0]
    c_experiment = data[:, component_index]

    reference = ReferenceIO(
        f'Peak {file_name}',
        time_experiment, c_experiment,
        component_system=ComponentSystem(["A"])
    )
    return reference


def create_comparator(reference):
    comparator = Comparator(name=f"Comp {reference.name}")
    comparator.add_reference(reference)
    comparator.add_difference_metric(
        'PeakPosition', reference,
        'outlet.outlet', components=["A"]
    )
    comparator.add_difference_metric(
        'PeakHeight', reference,
        'outlet.outlet', components=["A"]
    )
    return comparator


# %%
if __name__ == '__main__':
    create_in_silico_experimental_data()

    optimization_problem = OptimizationProblem('SMA_binding_parameters')

    simulator = Cadet()
    optimization_problem.add_evaluator(simulator)

    comparators = dict()

    for cv in [5, 30, 120]:
        process = create_process(cv)

        reference = load_reference(f"experimental_data/{cv}.csv")

        optimization_problem.add_evaluation_object(process)
        comparator = create_comparator(reference)

        # Here I stored the comparators for easy access from within the callback function.
        # Note, that the dictionary keys need to be identical to the process names
        # as defined in create_process() for this to work.
        comparators[str(cv)] = comparator

        optimization_problem.add_objective(
            comparator,
            name=f"Objective {comparator.name}",
            evaluation_objects=[process],  # limit this comparator to be applied to only this one process
            n_objectives=comparator.n_metrics,
            requires=[simulator]
        )

    optimization_problem.add_variable(
        name='adsorption_rate',
        parameter_path='flow_sheet.column.binding_model.adsorption_rate',
        lb=1e-3, ub=1e3,
        transform='auto',
        indices=[1]  # modify only the protein (component index 1) parameter
    )

    optimization_problem.add_variable(
        name='desorption_rate',
        parameter_path='flow_sheet.column.binding_model.desorption_rate',
        lb=1e-3, ub=1e3,
        transform='auto',
        indices=[1]  # modify only the protein (component index 1) parameter
    )

    optimization_problem.add_variable(
        name='equilibrium_constant',
        evaluation_objects=None,
        lb=1e-4, ub=1e3,
        transform='auto',
        indices=[1]  # modify only the protein (component index 1) parameter
    )

    optimization_problem.add_variable(
        name='kinetic_constant',
        evaluation_objects=None,
        lb=1e-4, ub=1e3,
        transform='auto',
        indices=[1]  # modify only the protein (component index 1) parameter
    )

    optimization_problem.add_variable(
        name='characteristic_charge',
        parameter_path='flow_sheet.column.binding_model.characteristic_charge',
        lb=1, ub=50,
        transform='auto',
        indices=[1]  # modify only the protein (component index 1) parameter
    )

    optimization_problem.add_variable_dependency(
        dependent_variable="desorption_rate",
        independent_variables=["kinetic_constant", ],
        transform=lambda k_kin: 1 / k_kin
    )

    optimization_problem.add_variable_dependency(
        dependent_variable="adsorption_rate",
        independent_variables=["kinetic_constant", "equilibrium_constant"],
        transform=lambda k_kin, k_eq: k_eq / k_kin
    )

    def callback(simulation_results, individual, evaluation_object, callbacks_dir='./'):
        comparator = comparators[evaluation_object.name]
        comparator.plot_comparison(
            simulation_results,
            file_name=f'{callbacks_dir}/{individual.id}_{evaluation_object}_comparison.png',
            show=False
        )

    optimization_problem.add_callback(callback, requires=[simulator])

    print(optimization_problem.variable_names)
    x0 = [1, 1, 1e-2, 1e-3, 10]
    ind = optimization_problem.create_individual(x0)
    optimization_problem.evaluate_callbacks(ind)

# %% [markdown]
# ```{note}
# For performance reasons, the optimization is currently not run when building the documentation.
# In future, we will try to sideload pre-computed results to also discuss them here.
# ```

# %%
# if __name__ == '__main__':
#     from CADETProcess.optimization import U_NSGA3
#
#     optimizer = U_NSGA3()
#     optimizer.n_cores = 4
#     optimizer.pop_size = 50
#     optimizer.n_max_gen = 5
#
#     optimization_results = optimizer.optimize(
#         optimization_problem,
#         use_checkpoint=False
#     )

# %%



import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from ax import (
    Arm,
    ComparisonOp,
    Data,
    Experiment,
    Metric,
    Models,
    MultiObjective,
    MultiObjectiveOptimizationConfig,
    Objective,
    OptimizationConfig,
    OutcomeConstraint,
    ParameterConstraint,
    ParameterType,
    RangeParameter,
    Runner,
    SearchSpace,
)
from ax.core.base_trial import BaseTrial
from ax.core.metric import MetricFetchE, MetricFetchResult
from ax.global_stopping.strategies.improvement import ImprovementGlobalStoppingStrategy
from ax.models.torch.botorch_defaults import get_qLogNEI
from ax.models.torch.botorch_modular.surrogate import Surrogate
from ax.service.utils.report_utils import exp_to_df
from ax.utils.common.result import Err, Ok
from botorch.acquisition.analytic import LogExpectedImprovement
from botorch.models.gp_regression import SingleTaskGP
from botorch.utils.sampling import manual_seed

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import Float, Typed, UnsignedInteger
from CADETProcess.optimization import OptimizerBase
from CADETProcess.optimization.optimizationProblem import OptimizationProblem
from CADETProcess.optimization.parallelizationBackend import (
    ParallelizationBackendBase,
    SequentialBackend,
)

__all__ = [
    "GPEI",
    "BotorchModular",
    "NEHVI",
    "qNParEGO",
]


class CADETProcessMetric(Metric):
    def __init__(
        self,
        name: str,
        lower_is_better: Union[bool, None] = None,
        properties: Union[Dict[str, Any], None] = None,
    ) -> None:
        super().__init__(name, lower_is_better, properties)

    def fetch_trial_data(
        self,
        trial: BaseTrial,
        **kwargs: Any,
    ) -> MetricFetchResult:
        try:
            trial_results = trial.run_metadata
            records = []
            for arm_name, arm in trial.arms_by_name.items():
                results_dict = {
                    "trial_index": trial.index,
                    "arm_name": arm_name,
                    "metric_name": self.name,
                }

                # this looks up the value of the objective function
                # generated in the runner
                arm_results = trial_results["arms"][arm]
                results_dict.update({"mean": arm_results[self.name]})
                results_dict.update({"sem": 0.0})

                records.append(results_dict)

            return Ok(Data(df=pd.DataFrame.from_records(records)))
        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )


class CADETProcessRunner(Runner):
    def __init__(
        self,
        optimization_problem: OptimizationProblem,
        parallelization_backend: ParallelizationBackendBase,
    ) -> None:
        self.optimization_problem = optimization_problem
        self.parallelization_backend = parallelization_backend

    @property
    def staging_required(self) -> bool:
        return False

    def run(self, trial: BaseTrial) -> Dict[str, Any]:
        # Get X from arms.
        X = []
        for arm in trial.arms:
            x = np.array(list(arm.parameters.values()))
            X.append(x)

        X = np.row_stack(X)

        # adjust the number of cores to the number of batch trials
        # See: https://github.com/fau-advanced-separations/CADET-Process/issues/53

        # Calculate objectives
        # Explore if adding a small amount of noise to the result helps BO
        objective_labels = self.optimization_problem.objective_labels
        obj_fun = self.optimization_problem.evaluate_objectives

        F = obj_fun(
            X,
            untransform=True,
            get_dependent_values=True,
            ensure_minimization=True,
            parallelization_backend=self.parallelization_backend,
        )

        # Calculate nonlinear constraints
        # Explore if adding a small amount of noise to the result helps BO
        if self.optimization_problem.n_nonlinear_constraints > 0:
            nonlincon_cv_fun = (
                self.optimization_problem.evaluate_nonlinear_constraints_violation
            )
            nonlincon_labels = self.optimization_problem.nonlinear_constraint_labels

            CV = nonlincon_cv_fun(
                X,
                untransform=True,
                get_dependent_values=True,
                parallelization_backend=self.parallelization_backend,
            )

        else:
            CV = None
            nonlincon_labels = None

        # Update trial information with results.
        trial_metadata = self.get_metadata(
            trial, F, objective_labels, CV, nonlincon_labels
        )

        return trial_metadata

    @staticmethod
    def get_metadata(
        trial: BaseTrial,
        F: np.ndarray,
        objective_labels: list[str],
        CV: np.ndarray,
        nonlincon_labels: list[str],
    ) -> dict:
        trial_metadata = {"name": str(trial.index)}
        trial_metadata.update({"arms": {}})

        for i, arm in enumerate(trial.arms):
            f_dict = {
                f"{metric}_axidx_{i_obj}": f_metric[i]
                for i_obj, (metric, f_metric) in enumerate(zip(objective_labels, F.T))
            }
            cv_dict = {}
            if CV is not None:
                cv_dict = {
                    f"{metric}_axidx_{i_constr}": cv_metric[i]
                    for i_constr, (metric, cv_metric) in enumerate(
                        zip(nonlincon_labels, CV.T)
                    )
                }
            trial_metadata["arms"].update({arm: {**f_dict, **cv_dict}})

        return trial_metadata


class AxInterface(OptimizerBase):
    """Wrapper around Ax's bayesian optimization API."""

    supports_bounds = True
    supports_multi_objective = False
    supports_linear_constraints = True
    supports_linear_equality_constraints = False
    supports_nonlinear_constraints = True

    early_stopping_improvement_window = UnsignedInteger(default=1000)
    early_stopping_improvement_bar = Float(default=1e-10)
    n_init_evals = UnsignedInteger(default=10)
    n_max_evals = UnsignedInteger(default=100)
    seed = UnsignedInteger(default=12345)

    _specific_options = [
        "n_init_evals",
        "n_max_evals",
        "seed",
        "early_stopping_improvement_window",
        "early_stopping_improvement_bar",
    ]

    @staticmethod
    def _setup_parameters(optimizationProblem: OptimizationProblem) -> list:
        parameters = []
        for var in optimizationProblem.independent_variables:
            lb, ub = var.transformed_bounds
            param = RangeParameter(
                name=var.name,
                parameter_type=ParameterType.FLOAT,
                lower=lb,
                upper=ub,
                log_scale=False,
                is_fidelity=False,
            )
            parameters.append(param)

        return parameters

    @staticmethod
    def _setup_linear_constraints(optimizationProblem: OptimizationProblem) -> list:
        A_transformed = optimizationProblem.A_independent_transformed
        b_transformed = optimizationProblem.b_transformed
        indep_vars = optimizationProblem.independent_variables
        parameter_constraints = []
        for a_t, b_t in zip(A_transformed, b_transformed):
            constr = ParameterConstraint(
                constraint_dict={var.name: a for var, a in zip(indep_vars, a_t)},
                bound=b_t,
            )
            parameter_constraints.append(constr)

        return parameter_constraints

    @classmethod
    def _setup_searchspace(
        cls, optimizationProblem: OptimizationProblem
    ) -> SearchSpace:
        return SearchSpace(
            parameters=cls._setup_parameters(optimizationProblem),
            parameter_constraints=cls._setup_linear_constraints(optimizationProblem),
        )

    def _setup_objectives(self) -> list:
        """Parse objective functions from optimization problem."""
        objective_names = self.optimization_problem.objective_labels

        objectives = []
        for i, obj_name in enumerate(objective_names):
            ax_metric = CADETProcessMetric(
                name=f"{obj_name}_axidx_{i}",
                lower_is_better=True,
            )

            obj = Objective(metric=ax_metric, minimize=True)
            objectives.append(obj)

        return objectives

    def _setup_outcome_constraints(self) -> list:
        """Parse nonliear constraint functions from optimization problem."""
        nonlincon_names = self.optimization_problem.nonlinear_constraint_labels

        outcome_constraints = []
        for i_constr, name in enumerate(nonlincon_names):
            ax_metric = CADETProcessMetric(name=f"{name}_axidx_{i_constr}")

            nonlincon = OutcomeConstraint(
                metric=ax_metric,
                op=ComparisonOp.LEQ,
                bound=0.0,
                relative=False,
            )
            outcome_constraints.append(nonlincon)

        return outcome_constraints

    def _create_manual_data(
        self, trial: BaseTrial, F: npt.ArrayLike, G: Optional[npt.ArrayLike] = None
    ) -> dict:
        objective_labels = self.optimization_problem.objective_labels
        nonlincon_labels = self.optimization_problem.nonlinear_constraint_labels
        return CADETProcessRunner.get_metadata(
            trial, F, objective_labels, G, nonlincon_labels
        )

    def _create_manual_trial(self, X: npt.ArrayLike) -> None:
        """Create trial from pre-evaluated data."""
        variables = self.optimization_problem.independent_variable_names

        for i, x in enumerate(X):
            trial = self.ax_experiment.new_trial()
            trial_data = {
                "input": {var: x_i for var, x_i in zip(variables, x)},
            }

            arm_name = f"{trial.index}_{0}"
            trial.add_arm(Arm(parameters=trial_data["input"], name=arm_name))
            trial.run()
            trial.mark_completed()
            self._post_processing(trial)

            # When returning to batch trials, the Arms can be initialized here
            # and then collectively returned. See commit history

    def _post_processing(self, trial: BaseTrial) -> None:
        """
        Run post processing.

        Ax holds the data of the model in a dataframe an experiment consists of trials
        which consist of arms in a sequential experiment, each trial only has one arm.

        Arms are evaluated. These hold the parameters.
        """
        op = self.optimization_problem

        # get the trial level data as a dataframe
        trial_data = self.ax_experiment.fetch_trials_data([trial.index])
        data = trial_data.df

        # DONE: Update for multi-processing. If n_cores > 1: len(arms) > 1 (oder @Flo?)
        X = np.array([list(arm.parameters.values()) for arm in trial.arms])
        objective_labels = [
            f"{obj_name}_axidx_{i}" for i, obj_name in enumerate(op.objective_labels)
        ]

        n_ind = len(X)

        # Get objective values
        F_data = data[data["metric_name"].isin(objective_labels)]
        assert np.all(
            F_data["metric_name"].values
            == np.repeat(objective_labels, len(X)).astype(object)
        )
        F = F_data["mean"].values.reshape((op.n_objectives, n_ind)).T

        # Get nonlinear constraint values
        if op.n_nonlinear_constraints > 0:
            nonlincon_labels = [
                f"{name}_axidx_{i}"
                for i, name in enumerate(op.nonlinear_constraint_labels)
            ]
            G_data = data[data["metric_name"].isin(nonlincon_labels)]
            assert np.all(
                G_data["metric_name"].values.tolist()
                == np.repeat(nonlincon_labels, len(X))
            )
            G = G_data["mean"].values.reshape((op.n_nonlinear_constraints, n_ind)).T

            nonlincon_cv_fun = op.evaluate_nonlinear_constraints_violation
            CV = nonlincon_cv_fun(X, untransform=True, get_dependent_values=True)
        else:
            G = None
            CV = None

        # Ideally, the current optimum w.r.t. single and multi objective can be
        # obtained at this point and passed to run_post_processing.
        # with X_opt_transformed. Implementation is pending.
        # See: https://github.com/fau-advanced-separations/CADET-Process/issues/53

        self.run_post_processing(
            X_transformed=X,
            F_minimized=F,
            G=G,
            CV_nonlincon=CV,
            current_generation=self.ax_experiment.num_trials,
            X_opt_transformed=None,
        )

    def _setup_model(self) -> None:
        """
        Initialize a pre-instantiated `Model` class.

        The class specifies the surrogate model (e.g. Gaussian Process) and acquisition function,
        which are used in the bayesian optimization algorithm.
        """
        raise NotImplementedError

        """Initialize an optimization configuration for Ax."""

    def _setup_optimization_config(
        self,
        objectives: list[Objective],
        outcome_constraints: OutcomeConstraint,
    ) -> None:
        raise NotImplementedError

    def _run(
        self, optimization_problem: OptimizationProblem, x0: npt.ArrayLike
    ) -> None:
        search_space = self._setup_searchspace(self.optimization_problem)
        objectives = self._setup_objectives()
        outcome_constraints = self._setup_outcome_constraints()
        optimization_config = self._setup_optimization_config(
            objectives=objectives, outcome_constraints=outcome_constraints
        )

        runner = CADETProcessRunner(
            optimization_problem=self.optimization_problem,
            parallelization_backend=SequentialBackend(),
        )

        self.global_stopping_strategy = ImprovementGlobalStoppingStrategy(
            min_trials=self.n_init_evals + self.early_stopping_improvement_window,
            window_size=self.early_stopping_improvement_window,
            improvement_bar=self.early_stopping_improvement_bar,
            inactive_when_pending_trials=True,
        )

        self.ax_experiment = Experiment(
            search_space=search_space,
            name=self.optimization_problem.name,
            optimization_config=optimization_config,
            runner=runner,
        )

        # Internal storage for tracking data
        self._data = self.ax_experiment.fetch_data()

        # Restore previous results from checkpoint
        if len(self.results.populations) > 0:
            for pop in self.results.populations:
                X, F, G = pop.x, pop.f, pop.g
                trial = self._create_manual_trial(X)
                trial.mark_running(no_runner_required=True)

                trial_data = self._create_manual_data(trial, F, G)
                trial.run_metadata.update(trial_data)
                trial.mark_completed()

        else:
            if x0 is not None:
                x0_init = np.array(x0, ndmin=2)

                if len(x0_init) < self.n_init_evals:
                    warnings.warn(
                        "Initial population smaller than popsize. "
                        "Creating missing entries."
                    )
                    n_remaining = self.n_init_evals - len(x0_init)
                    x0_remaining = optimization_problem.create_initial_values(
                        n_remaining, seed=self.seed, include_dependent_variables=False
                    )
                    x0_init = np.vstack((x0_init, x0_remaining))
                elif len(x0_init) > self.n_init_evals:
                    warnings.warn(
                        "Initial population larger than popsize. Omitting overhead."
                    )
                    x0_init = x0_init[0 : self.n_init_evals]

            else:
                # Create initial samples if they are not provided
                x0_init = self.optimization_problem.create_initial_values(
                    n_samples=self.n_init_evals,
                    include_dependent_variables=False,
                    seed=self.seed + 5641,
                )

            x0_init_transformed = np.array(optimization_problem.transform(x0_init))
            self._create_manual_trial(x0_init_transformed)
            print(exp_to_df(self.ax_experiment))

        n_iter = self.results.n_gen
        n_evals = self.results.n_evals

        global_stopping_message = None

        if n_evals >= self.n_max_evals:
            raise CADETProcessError(
                f"Initial number of evaluations exceeds `n_max_evals` "
                f"({self.n_max_evals})."
            )

        with manual_seed(seed=self.seed):
            while not (n_evals >= self.n_max_evals or n_iter >= self.n_max_iter):
                # Reinitialize GP+EI model at each step with updated data.
                modelbridge = self.train_model()

                print(f"Running optimization trial {n_evals + 1}/{self.n_max_evals}...")

                # samples can be accessed here by sample_generator.arms:
                sample_generator = modelbridge.gen(n=1)

                # A staging phase can be implemented here if needed.
                # See: https://github.com/fau-advanced-separations/CADET-Process/issues/53

                # The strategy itself will check if enough trials have already been
                # completed.
                (
                    stop_optimization,
                    global_stopping_message,
                ) = self.global_stopping_strategy.should_stop_optimization(
                    experiment=self.ax_experiment
                )

                if stop_optimization:
                    print(global_stopping_message)
                    break

                trial = self.ax_experiment.new_trial(generator_run=sample_generator)
                trial.run()

                trial.mark_completed()
                self._post_processing(trial)

                n_iter += 1
                n_evals += len(trial.arms)

        print(exp_to_df(self.ax_experiment))

        self.results.success = True
        self.results.exit_flag = 0
        self.results.exit_message = global_stopping_message


class SingleObjectiveAxInterface(AxInterface):
    def _setup_optimization_config(
        self,
        objectives: list[Objective],
        outcome_constraints: OutcomeConstraint,
    ) -> OptimizationConfig:
        return OptimizationConfig(
            objective=objectives[0], outcome_constraints=outcome_constraints
        )


class MultiObjectiveAxInterface(AxInterface):
    supports_multi_objective = True

    def _setup_optimization_config(
        self,
        objectives: list[Objective],
        outcome_constraints: OutcomeConstraint,
    ) -> MultiObjectiveOptimizationConfig:
        return MultiObjectiveOptimizationConfig(
            objective=MultiObjective(objectives),
            outcome_constraints=outcome_constraints,
        )


class GPEI(SingleObjectiveAxInterface):
    """Gaussian Process with Expected Improvement for single objectives."""

    def __repr__(self) -> str:
        """str: String representation of the optimization algorithm."""
        return "GPEI"

    def train_model(self) -> Models:
        """Train model."""
        return Models.GPEI(
            experiment=self.ax_experiment, data=self.ax_experiment.fetch_data()
        )


class BotorchModular(SingleObjectiveAxInterface):
    """
    Modular bayesian optimization algorithm.

    BotorchModular takes 2 optional arguments and uses the BOTORCH_MODULAR API of Ax to construct
    a Model which connects both components with the respective transforms necessary.

    Attributes
    ----------
    acquisition_fn: type, optional
        AcquisitionFunction class. The default is LogExpectedImprovement.
    surrogate_model: type, optional
        Model class. The default is SingleTaskGP.
    """

    acquisition_fn = Typed(ty=type, default=LogExpectedImprovement)
    surrogate_model = Typed(ty=type, default=SingleTaskGP)

    _specific_options = ["acquisition_fn", "surrogate_model"]

    def __repr__(self) -> str:
        """str: String representation of the optimization algorithm."""
        afn = self.acquisition_fn.__name__
        smn = self.surrogate_model.__name__

        return f"BotorchModular({smn}+{afn})"

    def train_model(self) -> NotImplementedError:
        """Train model."""
        raise NotImplementedError(
            "This model is currently broken. Please use Only GPEI or NEHVI"
        )
        return Models.BOTORCH_MODULAR(
            experiment=self.ax_experiment,
            surrogate=Surrogate(self.surrogate_model),
            botorch_acqf_class=self.acquisition_fn,
            data=self.ax_experiment.fetch_data(),
        )


class NEHVI(MultiObjectiveAxInterface):
    """Noisy expected hypervolume improvement multi-objective algorithm."""

    supports_single_objective = False

    def __repr__(self) -> str:
        """str: String representation of the optimization algorithm."""
        smn = "SingleTaskGP"
        afn = "NEHVI"

        return f"{smn}+{afn}"

    def train_model(self) -> Models:
        """Train model."""
        return Models.MOO(
            experiment=self.ax_experiment, data=self.ax_experiment.fetch_data()
        )


class qNParEGO(MultiObjectiveAxInterface):
    """
    qNParEGO multi-objective algorithm.

    ParEGO transforms the MOO problem into a single objective problem by applying a
    randomly weighted augmented Chebyshev scalarization to the objectives, and
    maximizing the expected improvement of that scalarized quantity (Knowles, 2006).
    Recently, Daulton et al. (2020) used a multi-output Gaussian process and
    compositional Monte Carlo objective to extend ParEGO to the batch setting (qParEGO),
    which proved to be a strong baseline for MOBO. Additionally, the authors proposed a
    noisy variant (qNParEGO), but the empirical evaluation of qNParEGO was limited.
    [Daulton et al. 2021 "Parallel Bayesian Optimization of Multiple Noisy Objectives
    with Expected Hypervolume Improvement"]
    """

    supports_single_objective = False

    def __repr__(self) -> str:
        """str: String representation of the algorithm."""
        smn = "SingleTaskGP"
        afn = "qNParEGO"

        return f"{smn}+{afn}"

    def train_model(self) -> Models:
        """Train model."""
        return Models.MOO(
            experiment=self.ax_experiment,
            data=self.ax_experiment.fetch_data(),
            acqf_constructor=get_qLogNEI,
            default_model_gen_options={
                "acquisition_function_kwargs": {
                    "chebyshev_scalarization": True,
                }
            },
        )



import functools as ft
import io
import os.path as op
import pickletools
import sqlite3
from typing import Any, Union

import dill as pickle
import diskcache
from diskcache.core import MODE_BINARY, MODE_PICKLE, MODE_RAW, MODE_TEXT, UNKNOWN

__all__ = ["DillDisk"]


class DillDisk(diskcache.Disk):
    """Cache key and value serialization for SQLite database and files."""

    def put(self, key: Any) -> None:
        """
        Convert `key` to fields key and raw for Cache table.

        :param key: key to convert
        :return: (database key, raw boolean) pair
        """
        # pylint: disable=unidiomatic-typecheck
        type_key = type(key)

        if type_key is bytes:
            return sqlite3.Binary(key), True
        elif (
            (type_key is str)
            or (type_key is int and -9223372036854775808 <= key <= 9223372036854775807)
            or (type_key is float)
        ):
            return key, True
        else:
            data = pickle.dumps(key, protocol=self.pickle_protocol)
            result = pickletools.optimize(data)
            return sqlite3.Binary(result), False

    def get(self, key: Any, raw: Any) -> Any:
        """
        Convert fields `key` and `raw` from Cache table to key.

        :param key: database key to convert
        :param bool raw: flag indicating raw database storage
        :return: corresponding Python key
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if raw:
            return bytes(key) if type(key) is sqlite3.Binary else key
        else:
            return pickle.load(io.BytesIO(key))

    def store(
        self,
        value: Any,
        read: bool,
        key: Any = UNKNOWN,
    ) -> tuple[int, int, Union[str, None], Union[Any, sqlite3.Binary]]:
        """
        Convert `value` to fields size, mode, filename, and value for Cache table.

        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Cache table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (type_value is int and -9223372036854775808 <= value <= 9223372036854775807)
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            else:
                filename, full_path = self.filename(key, value)
                self._write(full_path, io.BytesIO(value), "xb")
                return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.StringIO(value), "x", "UTF-8")
            size = op.getsize(full_path)
            return size, MODE_TEXT, filename, None
        elif read:
            reader = ft.partial(value.read, 2**22)
            filename, full_path = self.filename(key, value)
            iterator = iter(reader, b"")
            size = self._write(full_path, iterator, "xb")
            return size, MODE_BINARY, filename, None
        else:
            result = pickle.dumps(value, protocol=self.pickle_protocol)

            if len(result) < min_file_size:
                return 0, MODE_PICKLE, None, sqlite3.Binary(result)
            else:
                filename, full_path = self.filename(key, value)
                self._write(full_path, io.BytesIO(result), "xb")
                return len(result), MODE_PICKLE, filename, None

    def fetch(self, mode: int, filename: str, value: Any, read: bool) -> Any:
        """
        Convert fields `mode`, `filename`, and `value` from Cache table to value.

        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        :raises: IOError if the value cannot be read
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck,consider-using-with
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read:
                return open(op.join(self._directory, filename), "rb")
            else:
                with open(op.join(self._directory, filename), "rb") as reader:
                    return reader.read()
        elif mode == MODE_TEXT:
            full_path = op.join(self._directory, filename)
            with open(full_path, "r", encoding="UTF-8") as reader:
                return reader.read()
        elif mode == MODE_PICKLE:
            if value is None:
                with open(op.join(self._directory, filename), "rb") as reader:
                    return pickle.load(reader)
            else:
                return pickle.load(io.BytesIO(value))



import unittest

import numpy as np
from CADETProcess.dataStructure import (
    Aggregator,
    Callable,
    Constant,
    DependentlyModulatedUnsignedList,
    Float,
    FloatList,
    Integer,
    IntegerList,
    List,
    Matrix,
    NdPolynomial,
    Polynomial,
    RangedFloat,
    SizedAggregator,
    SizedList,
    SizedNdArray,
    SizedUnsignedList,
    SizedUnsignedNdArray,
    String,
    Structure,
    Switch,
    Typed,
    UnsignedInteger,
    Vector,
)


class TestDescription(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            param_with_description = Integer(description="foo")

        self.model = Model()

    def test_description(self):
        self.assertEqual(type(self.model).param_with_description.description, "foo")

    def test_modified_descriptor(self):
        type(self.model).param_with_description.description = "bar"
        self.assertEqual(type(self.model).param_with_description.description, "bar")


class TestParameterDictionaries(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            param = Integer()
            param_default = Integer(default=1)
            no_param = None

            _parameters = ["param", "param_default"]

        self.model = Model()

    def test_parameters_dict_getter(self):
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": None, "param_default": 1}
        )

        self.model.param = 1
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": 1, "param_default": 1}
        )

        self.model.param = None
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": None, "param_default": 1}
        )

        self.model.param_default = 2
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": None, "param_default": 2}
        )

        self.model.param_default = None
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": None, "param_default": 1}
        )

    def test_parameters_dict_setter(self):
        self.model.parameters = {"param": 1, "param_default": 2}
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": 1, "param_default": 2}
        )

        self.model.parameters = {"param": 2}
        np.testing.assert_equal(
            self.model._parameters_dict, {"param": 2, "param_default": 2}
        )

        with self.assertRaises(ValueError):
            self.model.parameters = {"not_a_valid_param": 1}


class TestConstant(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            const_int = Constant(value=0)
            const_list = Constant(value=[0, 1])

        self.model = Model()

    def test_constant(self):
        self.assertEqual(self.model.const_int, 0)

        with self.assertRaises(ValueError):
            self.model.const_int = 1

        self.assertEqual(self.model.const_list, [0, 1])

        with self.assertRaises(ValueError):
            self.model.const_list = [1, 2]

    def test_error_when_no_value(self):
        with self.assertRaises(TypeError):

            class NoValue(Structure):
                const = Constant()


class TestSwitch(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            switch = Switch(valid=["foo", "bar"], default="foo")

        self.model = Model()

    def test_value(self):
        self.model.switch = "bar"
        self.assertEqual(self.model.switch, "bar")

        with self.assertRaises(ValueError):
            self.model.switch = "spam"

    def test_default(self):
        self.assertEqual(self.model.switch, "foo")

        with self.assertRaises(ValueError):

            class InvalidDefault(Structure):
                switch = Switch(valid=["foo", "bar"], default="spam")


class TestTyped(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            string_param = String(default="foo")
            list_param = List(default=[1, 2])
            dynamic_type_param = Typed(ty=list, default=[0, 1])

        self.model = Model()
        self.model_other = Model()

    def test_values(self):
        self.model.string_param = "string_param"
        self.assertEqual(self.model.string_param, "string_param")

        with self.assertRaises(TypeError):
            self.model.string_param = 0

        self.model.list_param = [
            0,
        ]
        self.assertEqual(
            self.model.list_param,
            [
                0,
            ],
        )

        with self.assertRaises(TypeError):
            self.model.list_param = 0

        self.model.dynamic_type_param = [0]
        with self.assertRaises(TypeError):
            self.model.dynamic_type_param = 0

    def test_default(self):
        self.assertEqual(self.model.string_param, "foo")
        self.assertEqual(self.model.list_param, [1, 2])
        self.assertEqual(self.model.dynamic_type_param, [0, 1])

        self.assertFalse(self.model.list_param is self.model_other.list_param)
        self.assertFalse(self.model.dynamic_type_param is self.model_other.list_param)

        with self.assertRaises(TypeError):

            class WrongDefaultType(Structure):
                param_1 = List(default="string")

        with self.assertRaises(TypeError):

            class WrongDefaultType(Structure):
                param_1 = List(default=1)

            model = WrongDefaultType()
            model.param_1


class TestCallable(unittest.TestCase):
    def setUp(self):
        def default_method(x):
            return x

        class Model(Structure):
            method = Callable()
            method_default = Callable(default=default_method)

        self.model = Model()

    def test_values(self):
        def method(x):
            return 2 * x

        self.model.method = method
        assert callable(self.model.method)

        self.assertEqual(self.model.method(2), 4)

        with self.assertRaises(TypeError):
            self.model.method = 2

    def test_default(self):
        assert callable(self.model.method_default)
        self.assertEqual(self.model.method_default(2), 2)


class TestDtype(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            integer_list_param = IntegerList()
            float_list_param = FloatList()

        self.model = Model()
        self.model_other = Model()

    def test_values(self):
        self.model.integer_list_param = [1, 2]
        self.assertEqual(self.model.integer_list_param, [1, 2])

        with self.assertRaises(ValueError):
            self.model.integer_list_param = [1, "foo"]

        self.model.float_list_param = [1.0, 2]
        self.assertEqual(self.model.float_list_param, [1.0, 2.0])


class TestRanged(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            bound_float_param = RangedFloat(lb=-1, ub=1, default=0)
            unsigned_integer_param = UnsignedInteger(default=1)

        self.model = Model()

    def test_values(self):
        self.model.bound_float_param = 0.5
        self.assertEqual(self.model.bound_float_param, 0.5)

        with self.assertRaises(ValueError):
            self.model.bound_float_param = -2

        with self.assertRaises(ValueError):
            self.model.bound_float_param = 2

        self.model.unsigned_integer_param = 10
        with self.assertRaises(ValueError):
            self.model.unsigned_integer_param = -2

    def test_default(self):
        self.assertEqual(self.model.bound_float_param, 0.0)
        self.assertEqual(self.model.unsigned_integer_param, 1)

        with self.assertRaises(ValueError):

            class InvalidDefault(Structure):
                param_1 = UnsignedInteger(default=-1)


class TestSizedUnified(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            sized_list = SizedList(size=4, default=0)
            sized_array = SizedNdArray(size=(2, 4), default=0)

            sized_unsigned_list = SizedUnsignedList(size=4, default=1)
            sized_unsigned_array = SizedUnsignedNdArray(size=(2, 4), default=1)

            sized_list_full_default = SizedList(size=4, default=[1, 2, 3, 4])
            sized_array_full_default = SizedNdArray(
                size=(2, 4), default=[[1, 2, 3, 4], [5, 6, 7, 8]]
            )

        self.model = Model()

    def test_values(self):
        self.model.sized_list = [1, 2, 3, 4]
        np.testing.assert_equal(self.model.sized_list, [1, 2, 3, 4])

        self.model.sized_array = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        np.testing.assert_equal(self.model.sized_array, [[1, 2, 3, 4], [5, 6, 7, 8]])

        self.model.sized_array = [[0, 1, 2, 3], [4, 5, 6, 7]]
        np.testing.assert_equal(self.model.sized_array, [[0, 1, 2, 3], [4, 5, 6, 7]])

        self.model.sized_unsigned_list = [1, 2, 3, 4]
        np.testing.assert_equal(self.model.sized_unsigned_list, [1, 2, 3, 4])

        with self.assertRaises(ValueError):
            self.model.sized_unsigned_list = [-1, 2, 3, 4]

        self.model.sized_unsigned_array = [[1, 2, 3, 4], [5, 6, 7, 8]]
        np.testing.assert_equal(
            self.model.sized_unsigned_array, [[1, 2, 3, 4], [5, 6, 7, 8]]
        )

        with self.assertRaises(ValueError):
            self.model.sized_unsigned_array = [[-1, 2, 3, 4], [5, 6, 7, 8]]

    def test_default(self):
        np.testing.assert_equal(self.model.sized_list, [0, 0, 0, 0])
        np.testing.assert_equal(self.model.sized_array, [[0, 0, 0, 0], [0, 0, 0, 0]])

        np.testing.assert_equal(self.model.sized_unsigned_list, [1, 1, 1, 1])
        np.testing.assert_equal(
            self.model.sized_unsigned_array, [[1, 1, 1, 1], [1, 1, 1, 1]]
        )

        # Full default
        # List
        np.testing.assert_equal(self.model.sized_list_full_default, [1, 2, 3, 4])

        with self.assertRaises(ValueError):

            class InvalidDefaultSize(Structure):
                param_1 = SizedList(size=4, default=[1, 2, 3])

        # Array
        np.testing.assert_equal(
            self.model.sized_array_full_default, [[1, 2, 3, 4], [5, 6, 7, 8]]
        )
        with self.assertRaises(ValueError):

            class InvalidDefaultSize(Structure):
                param_1 = SizedNdArray(
                    size=(4, 2), default=[[1, 2, 3, 4], [5, 6, 7, 8], [10, 11, 12, 13]]
                )

        # Invalid range
        with self.assertRaises(ValueError):

            class InvalidDefaultRange(Structure):
                param_1 = SizedUnsignedList(size=4, default=-1)


class TestSizedDependent(unittest.TestCase):
    """Previous test methods for dependently sized parameters."""

    def setUp(self):
        class Model(Structure):
            dep_1 = UnsignedInteger()
            dep_2 = UnsignedInteger(default=3)

            list_single_dep_param = SizedUnsignedList(size="dep_1", default=1)
            array_single_dep_param = SizedUnsignedNdArray(size="dep_1", default=1)

            list_double_dep_param = SizedUnsignedList(
                size=("dep_1", "dep_2"), default=1
            )
            array_double_dep_param = SizedUnsignedNdArray(
                size=("dep_1", "dep_2"), default=1
            )

            list_double_dep_with_int = SizedUnsignedList(size=("dep_1", 2), default=2)
            array_double_dep_with_int = SizedUnsignedNdArray(
                size=("dep_1", 2), default=2
            )

        self.model = Model()

    def test_values(self):
        # Missing dependency value
        with self.assertRaises(ValueError):
            self.model.list_single_dep_param = [2, 2]

        with self.assertRaises(ValueError):
            self.model.array_single_dep_param = [2, 2]

        self.model.dep_1 = 2

        # Single dependency
        self.model.list_single_dep_param = [2, 2]
        np.testing.assert_equal(self.model.list_single_dep_param, [2, 2])

        self.model.array_single_dep_param = [2, 2]
        np.testing.assert_equal(self.model.array_single_dep_param, [2, 2])

        # Wrong shape
        with self.assertRaises(ValueError):
            self.model.list_single_dep_param = [2, 2, 2]

        # Multiple dependencies
        np.testing.assert_array_equal(self.model.list_double_dep_param, np.ones((6,)))
        np.testing.assert_array_equal(
            self.model.array_double_dep_param, np.ones((2, 3))
        )

        # Size error
        with self.assertRaises(ValueError):
            self.model.list_double_dep_param = np.ones((3, 3)).flatten().tolist()
        with self.assertRaises(ValueError):
            self.model.array_double_dep_param = np.ones((3, 3))

        # Range error
        with self.assertRaises(ValueError):
            self.model.list_double_dep_param = (-1 * np.ones((2, 3))).flatten().tolist()
        with self.assertRaises(ValueError):
            self.model.array_double_dep_param = -1 * np.ones((2, 3))

        # Dependencies with integer dimension
        np.testing.assert_equal(self.model.list_double_dep_with_int, [2, 2, 2, 2])
        np.testing.assert_equal(self.model.array_double_dep_with_int, [[2, 2], [2, 2]])

    def test_default(self):
        self.model.dep_1 = 2

        np.testing.assert_equal(self.model.list_single_dep_param, [1, 1])
        np.testing.assert_equal(self.model.array_single_dep_param, [1, 1])

        # Full default
        with self.assertRaises(ValueError):

            class InvalidDefaultSizeList(Structure):
                dep_1 = UnsignedInteger()
                param_1 = SizedUnsignedList(size="dep_1", default=[1, 2, 3])

        with self.assertRaises(ValueError):

            class InvalidDefaultSizeArray(Structure):
                param_1 = SizedUnsignedNdArray(
                    size=(4, "dep_1"), default=[[1, 2, 3, 4], [5, 6, 7, 8]]
                )

        # Range error
        with self.assertRaises(ValueError):

            class InvalidDefault(Structure):
                list_single_dep_param = SizedUnsignedList(size="dep_1", default=-1)


class TestPolynomial(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            poly_param = Polynomial(n_coeff=2, default=0)
            ndpoly_param = NdPolynomial(n_entries=2, n_coeff=4)

            n_coeff = 4
            poly_param_dep = Polynomial(size=("n_coeff"))

            n_entries = 3
            ndpoly_param_dep = NdPolynomial(size=("n_entries", "n_coeff"))

        self.model = Model()

    def test_parameter(self):
        with self.assertRaises(ValueError):

            class TestDuplicateCoeffs(Structure):
                n_coeff = 2
                faulty = Polynomial(n_coeff=2, size=("n_coeff"))

        with self.assertRaises(ValueError):

            class TestMissingEntries(Structure):
                n_coeff = 2
                faulty = NdPolynomial(size=("n_coeff"))

        with self.assertRaises(ValueError):

            class TestDuplicateCoeff(Structure):
                n_coeff = 2
                entries = 2
                faulty = NdPolynomial(n_entries=2, n_coeff=2, size=("n_coeff"))

        with self.assertRaises(ValueError):

            class TestInvalidDefault(Structure):
                faulty = NdPolynomial(n_entries=2, n_coeff=2, default=1)

    def test_values(self):
        # Single entry n_coeff
        self.model.poly_param = 2
        np.testing.assert_equal(self.model.poly_param, [2, 0])

        self.model.poly_param = [3, 2]
        np.testing.assert_equal(self.model.poly_param, [3, 2])

        with self.assertRaises(ValueError):
            self.model.poly_param = [3, 2, 1]

        # Single entry with dependency
        self.model.n_coeff = 3

        with self.assertRaises(ValueError):
            self.model.poly_param_dep = [1, 2, 3, 4]

        self.model.poly_param_dep = [1, 2, 3]
        np.testing.assert_equal(self.model.poly_param_dep, [1, 2, 3])

        self.model.n_coeff = 4
        self.model.poly_param_dep = [1, 2, 3, 4]
        np.testing.assert_equal(self.model.poly_param_dep, [1, 2, 3, 4])

        # Multiple entries
        self.model.ndpoly_param = 2
        np.testing.assert_equal(self.model.ndpoly_param, [[2, 0, 0, 0], [2, 0, 0, 0]])

        self.model.ndpoly_param = [3, 2]
        np.testing.assert_equal(self.model.ndpoly_param, [[3, 0, 0, 0], [2, 0, 0, 0]])

        self.model.ndpoly_param = [[1, 2], [0, 1, 2]]
        np.testing.assert_equal(
            self.model.ndpoly_param, [[1, 2, 0.0, 0.0], [0, 1, 2, 0]]
        )

        with self.assertRaises(ValueError):
            self.model.ndpoly_param = [3, 2, 1]

        with self.assertRaises(ValueError):
            self.model.ndpoly_param = [[1], [1], [1]]

        # Multiple entries with dependency
        self.model.n_coeff = 3

        with self.assertRaises(ValueError):
            self.model.ndpoly_param_dep = [[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]

        self.model.ndpoly_param_dep = 2
        np.testing.assert_equal(
            self.model.ndpoly_param_dep, [[2, 0, 0], [2, 0, 0], [2, 0, 0]]
        )
        self.model.n_coeff = 4

        self.model.ndpoly_param_dep = [[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]
        np.testing.assert_equal(
            self.model.ndpoly_param_dep, [[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]
        )

        self.model.ndpoly_param_dep = [1, 2, 3]
        np.testing.assert_equal(
            self.model.ndpoly_param_dep, [[1, 0, 0, 0], [2, 0, 0, 0], [3, 0, 0, 0]]
        )

        self.model.ndpoly_param_dep = [[1, 2], [0, 1, 2], [0]]
        np.testing.assert_equal(
            self.model.ndpoly_param_dep, [[1, 2, 0, 0], [0, 1, 2, 0], [0, 0, 0, 0]]
        )

        with self.assertRaises(ValueError):
            self.model.ndpoly_param_dep = [1, 2, 3, 4]

        with self.assertRaises(ValueError):
            self.model.ndpoly_param_dep = [[1], [1], [1], [1]]

    def test_default(self):
        """
        Currently, only `0` is allowed as default value.

        Notes
        -----
        Technically, default values would be possible from point of the NdPolynomial
        class. However, due to its use as an Event Parameter, this is currently diabled.
        """
        np.testing.assert_equal(self.model.poly_param, [0, 0])

        np.testing.assert_equal(self.model.ndpoly_param, None)


class TestModulated(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            n_mod = 4

            modulated_list = DependentlyModulatedUnsignedList(size="n_mod")

        self.model = Model()

    def test_value(self):
        self.model.modulated_list = [1, 2, 3, 4]

        np.testing.assert_equal(self.model.modulated_list, [1, 2, 3, 4])

        self.model.modulated_list = [1, 2, 3, 4, 5, 6, 7, 8]

        with self.assertRaises(ValueError):
            self.model.modulated_list = [1, 2, 3, 4, 5]

        self.model.n_mod = 5
        with self.assertRaises(ValueError):
            self.model.modulated_list


class TestDimensionalized(unittest.TestCase):
    def setUp(self):
        class Model(Structure):
            vector = Vector()
            matrix = Matrix()

        self.model = Model()

    def test_value(self):
        self.model.vector = [1, 2, 3]

        with self.assertRaises(ValueError):
            self.model.vector = [[1, 2, 3], [4, 5, 6]]

        with self.assertRaises(ValueError):
            self.model.matrix = [1, 2]


class TestAggregator(unittest.TestCase):
    def setUp(self):
        class DummyInstance(Structure):
            float_param = Float(default=1.0)
            sized_param = SizedNdArray(size=4)
            sized_param_transposed = SizedNdArray(size=2)

        class Model(Structure):
            aggregator = Aggregator("float_param", "container")
            sized_aggregator = SizedAggregator("sized_param", "container")
            transposed_sized_aggregator = SizedAggregator(
                "sized_param_transposed", "container", transpose=True
            )

            def __init__(self):
                self.container = [
                    DummyInstance(
                        float_param=i,
                        sized_param=[float(i * j) for j in range(4)],
                        sized_param_transposed=[float(i * j) for j in range(2)],
                    )
                    for i in range(3)
                ]

        self.model = Model()

    def test_value(self):
        # Aggregator
        self.assertAlmostEqual(self.model.aggregator, [0.0, 1.0, 2.0])

        new_value = [1, 2, 3]
        self.model.aggregator = new_value

        self.assertAlmostEqual(self.model.aggregator, new_value)
        for con, val in zip(self.model.container, new_value):
            self.assertAlmostEqual(con.float_param, val)

        # Test setting incorrect types or dimensions:
        with self.assertRaises((ValueError, TypeError)):
            self.model.aggregator = 3

        # Test setting slices and indexes
        self.model.aggregator[0] = 3.0
        self.assertAlmostEqual(self.model.aggregator[0], 3.0)

        # SizedAggregator
        np.testing.assert_almost_equal(
            self.model.sized_aggregator,
            [
                [0, 0, 0, 0],
                [0, 1, 2, 3],
                [0, 2, 4, 6],
            ],
        )

        new_value = [
            [1.0, 2.0, 3.0, 4.0],
            [2.0, 3.0, 4.0, 5.0],
            [3.0, 4.0, 5.0, 6.0],
        ]
        self.model.sized_aggregator = new_value

        np.testing.assert_almost_equal(self.model.sized_aggregator, new_value)
        for con, val in zip(self.model.container, new_value):
            np.testing.assert_almost_equal(con.sized_param, val)

        # Test setting incorrect types or dimensions:
        with self.assertRaises((ValueError, TypeError)):
            self.model.sized_aggregator = 3

        with self.assertRaises((ValueError, TypeError)):
            self.model.sized_aggregator = [1.0, 2.0, 3.0, 4.0]

        with self.assertRaises(IndexError):
            self.model.sized_aggregator[3] = 3.0

        # Test setting slices and indexes
        new_value = [
            [1.5, 2.5, 3.5, 4.5],
            [2.5, 3.5, 4.5, 5.5],
            [3.5, 4.5, 5.5, 6.5],
        ]

        self.model.sized_aggregator[0] = new_value[0]
        np.testing.assert_almost_equal(self.model.sized_aggregator[0], new_value[0])

        self.model.sized_aggregator[0, 0] = 7.5
        np.testing.assert_almost_equal(self.model.sized_aggregator[0, 0], 7.5)

        # Transposed SizedAggregator
        np.testing.assert_almost_equal(
            self.model.transposed_sized_aggregator,
            [
                [0, 0, 0],
                [0, 1, 2],
            ],
        )
        new_value = [
            [1, 2, 3],
            [2, 3, 4],
        ]
        self.model.transposed_sized_aggregator = new_value

        np.testing.assert_almost_equal(
            self.model.transposed_sized_aggregator, new_value
        )
        for con, val in zip(self.model.container, np.array(new_value).T):
            np.testing.assert_almost_equal(con.sized_param_transposed, val)


if __name__ == "__main__":
    unittest.main()



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# (lwe_example_concentration)=
# # Concentration Gradients
#
# ```{figure} ./figures/flow_sheet_concentration.svg
# Flow sheet for load-wash-elute process using a single inlet.
# ```

# %%
import numpy as np

import matplotlib
matplotlib.use("TkAgg")

from CADETProcess.processModel import ComponentSystem, HICConstantWaterActivity
from CADETProcess.processModel import Inlet, GeneralRateModel, Outlet
from CADETProcess.processModel import FlowSheet
from CADETProcess.processModel import Process

# Component System
component_system = ComponentSystem()
component_system.add_component('Salt')
component_system.add_component('A')
component_system.add_component('B')
component_system.add_component('C')

# Binding Model
binding_model = HICConstantWaterActivity(component_system, name='HIC_SWA')
# binding_model = HICWaterOnHydrophobicSurfaces(component_system, name='HIC_WHS')
binding_model.is_kinetic = True
binding_model.adsorption_rate = [0.0, 0.7, 1, 200]
binding_model.desorption_rate = [0.0, 1000, 1000, 1000]
binding_model.hic_characteristic = [0.0, 10, 13, 4]
binding_model.capacity = [0.0, 10000000, 10000000, 10000000]
binding_model.beta_0 = 10. ** -0.5
binding_model.beta_1 = 10. ** -3.65
binding_model.bound_states = [0, 1, 1, 1]

# Unit Operations
inlet = Inlet(component_system, name='inlet')
inlet.flow_rate = 6.683738370512285e-8

column = GeneralRateModel(component_system, name='column')
column.binding_model = binding_model

column.length = 0.014
column.diameter = 0.02
column.bed_porosity = 0.37
column.particle_radius = 4.5e-5
column.particle_porosity = 0.75
column.axial_dispersion = 5.75e-8
column.film_diffusion = column.n_comp * [6.9e-6]
column.pore_diffusion = [7e-10, 6.07e-11, 6.07e-11, 6.07e-11]
column.surface_diffusion = column.n_bound_states * [0.0]

salt_gradient_start_concentration = 3000
salt_gradient_end_concentration = 50

column.c = [salt_gradient_start_concentration, 0, 0, 0]
column.cp = [salt_gradient_start_concentration, 0, 0, 0]
column.q = [0, 0, 0]

outlet = Outlet(component_system, name='outlet')

# Flow Sheet
flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(inlet)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet, product_outlet=True)

flow_sheet.add_connection(inlet, column)
flow_sheet.add_connection(column, outlet)

# %% [markdown]
# ```{figure} ./figures/events_concentration.svg
# Events of load-wash-elute process using a single inlet and modifying its concentration.
# ```

# %%
# Process
process = Process(flow_sheet, 'lwe')
process.cycle_time = 2000.0

load_duration = 9
t_gradient_start = 90.0
gradient_duration = process.cycle_time - t_gradient_start

c_load = np.array([salt_gradient_start_concentration, 1.0, 1.0, 1.0])
c_wash = np.array([salt_gradient_start_concentration, 0.0, 0.0, 0.0])
c_elute = np.array([salt_gradient_end_concentration, 0.0, 0.0, 0.0])
gradient_slope = (c_elute - c_wash) / gradient_duration
c_gradient_poly = np.array(list(zip(c_wash, gradient_slope)))

process.add_event('load', 'flow_sheet.inlet.c', c_load)
process.add_event('wash', 'flow_sheet.inlet.c', c_wash, load_duration)
process.add_event('grad_start', 'flow_sheet.inlet.c', c_gradient_poly, t_gradient_start)

# %%
if __name__ == '__main__':
    from CADETProcess.simulator import Cadet

    process_simulator = Cadet()

    simulation_results = process_simulator.simulate(process)

    from CADETProcess.plotting import SecondaryAxis

    sec = SecondaryAxis()
    sec.components = ['Salt']
    sec.y_label = '$c_{salt}$'

    simulation_results.solution.column.outlet.plot(secondary_axis=sec)



import warnings
from typing import Any, Optional

import numpy as np
import numpy.typing as npt
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.moo.unsga3 import UNSGA3
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.core.repair import Repair
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.ref_dirs import get_reference_directions

from CADETProcess.dataStructure import UnsignedFloat, UnsignedInteger
from CADETProcess.optimization import (
    OptimizationProblem,
    OptimizerBase,
    ParallelizationBackendBase,
)


class PymooInterface(OptimizerBase):
    """Wrapper around pymoo."""

    is_population_based = True

    supports_multi_objective = True
    supports_linear_constraints = True
    supports_linear_equality_constraints = True
    supports_nonlinear_constraints = True
    supports_bounds = True

    ignore_linear_constraints_config = True

    seed = UnsignedInteger(default=12345)
    pop_size = UnsignedInteger()

    xtol = UnsignedFloat(default=1e-8)
    ftol = UnsignedFloat(default=0.0025)
    cvtol = UnsignedFloat(default=1e-6)

    n_max_gen = UnsignedInteger()
    n_skip = UnsignedInteger(default=0)

    x_tol = xtol                # Alias for uniform interface
    f_tol = ftol                # Alias for uniform interface
    cv_nonlincon_tol = cvtol    # Alias for uniform interface
    n_max_iter = n_max_gen      # Alias for uniform interface

    _specific_options = [
        "seed",
        "pop_size",
        "xtol",
        "ftol",
        "cvtol",
        "n_max_gen",
        "n_skip",
    ]

    def _run(self, optimization_problem: OptimizationProblem, x0: Optional[list] = None) -> None:
        """
        Solve optimization problem using functional pymoo implementation.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            DESCRIPTION.
        x0 : list, optional
            Initial population of independent variables in untransformed space.

        Returns
        -------
        results : OptimizationResults
            Optimization results including optimization_problem and solver
            configuration.

        See Also
        --------
        evaluate_objectives
        options
        """
        pop_size = self.get_population_size(optimization_problem)

        if x0 is not None:
            pop = x0
        else:
            pop = optimization_problem.create_initial_values(
                pop_size, seed=self.seed, include_dependent_variables=False
            )

        pop = np.array(pop, ndmin=2)

        if len(pop) < pop_size:
            warnings.warn(
                "Initial population smaller than popsize. Creating missing entries."
            )
            n_remaining = pop_size - len(pop)
            remaining = optimization_problem.create_initial_values(
                n_remaining, seed=self.seed, include_dependent_variables=False
            )
            pop = np.vstack((pop, remaining))
        elif len(pop) > pop_size:
            warnings.warn("Initial population larger than popsize. Omitting overhead.")
            pop = pop[0:pop_size]

        pop = np.array(optimization_problem.transform(pop))

        problem = PymooProblem(optimization_problem, self.parallelization_backend)

        ref_dirs = get_reference_directions(
            "energy",
            optimization_problem.n_objectives,
            pop_size,
            seed=1,
        )

        algorithm = self._cls(
            ref_dirs=ref_dirs,
            pop_size=pop_size,
            sampling=pop,
            repair=RepairIndividuals(self, optimization_problem),
        )

        n_max_gen = self.get_max_number_of_generations(optimization_problem)

        termination = DefaultMultiObjectiveTermination(
            xtol=self.xtol,
            cvtol=self.cvtol,
            ftol=self.ftol,
            n_max_gen=n_max_gen,
            n_max_evals=self.n_max_evals,
            n_skip=self.n_skip,
        )

        algorithm.setup(
            problem,
            termination=termination,
            seed=self.seed,
            verbose=True,
            save_history=False,
            output=MultiObjectiveOutput(),
        )

        # Restore previous results from checkpoint
        for pop in self.results.populations:
            _ = algorithm.ask()
            if optimization_problem.n_nonlinear_constraints > 0:
                pop = Population.new("X", pop.x, "F", pop.f, "G", pop.cv)
                pop.apply(lambda ind: ind.evaluated.update({"F", "G"}))
                algorithm.evaluator.eval(problem, pop, evaluate_values_of=["F", "G"])
            else:
                pop = Population.new("X", pop.x, "F", pop.f)
                pop.apply(lambda ind: ind.evaluated.update({"F"}))
                algorithm.evaluator.eval(problem, pop, evaluate_values_of=["F"])
            algorithm.evaluator.n_eval += len(pop)
            algorithm.tell(infills=pop)

        while algorithm.has_next():
            # Get current generation
            pop = algorithm.ask()
            X = pop.get("X").tolist()

            # Evaluate objectives and report results
            algorithm.evaluator.eval(problem, pop)

            F = pop.get("F").tolist()
            if optimization_problem.n_nonlinear_constraints > 0:
                G = pop.get("CADET_G").tolist()
                CV = pop.get("CADET_CV").tolist()
            else:
                G = None
                CV = None

            # Handle issue of pymoo not handling np.inf
            pop.set("F", np.nan_to_num(F, posinf=1e300))

            algorithm.tell(infills=pop)

            # Post generation processing
            X_opt = algorithm.opt.get("X").tolist()
            self.run_post_processing(X, F, G, CV, algorithm.n_gen - 1, X_opt)

        if algorithm.n_gen >= n_max_gen:
            success = True
            exit_flag = 1
            exit_message = "Max number of generations exceeded."
        else:
            success = True
            exit_flag = 0
            exit_message = "Success"

        self.results.success = success
        self.results.exit_flag = exit_flag
        self.results.exit_message = exit_message

        return self.results

    def get_population_size(
            self,
            optimization_problem: OptimizationProblem,
    ) -> int:
        """
        Determine the population size for an optimization problem.

        This method calculates the population size based on the number of independent
        variables in the optimization problem. If `pop_size` is not set, it defaults to
        the minimum of 400 or the maximum of 50 times the number of independent variables
        and 50.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            The optimization problem for which to determine the population size.

        Returns
        -------
        int
            The population size.
        """
        if self.pop_size is None:
            return min(400, max(50 * optimization_problem.n_independent_variables, 50))
        else:
            return self.pop_size

    def get_max_number_of_generations(
            self,
            optimization_problem: OptimizationProblem,
    ) -> int:
        """
        Determine the maximum number of generations for an optimization problem.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            The optimization problem for which to determine the maximum number of generations.

        Returns
        -------
        int
            The maximum number of generations.
        """
        if self.n_max_gen is None:
            return min(100, max(10 * optimization_problem.n_independent_variables, 40))
        else:
            return self.n_max_gen


class NSGA2(PymooInterface):
    """NSGA2 Algorithm."""

    _cls = NSGA2

    def __str__(self) -> str:
        """str: String representation."""
        return "NSGA2"


class U_NSGA3(PymooInterface):
    """U-NSGA3 Algorithm."""

    _cls = UNSGA3

    def __str__(self) -> str:
        """str: String representation."""
        return "UNSGA3"


class PymooProblem(Problem):
    """Class to implement Pymoo Problem interface."""

    def __init__(
        self,
        optimization_problem: OptimizationProblem,
        parallelization_backend: ParallelizationBackendBase,
        **kwargs: Any,
    ) -> None:
        self.optimization_problem = optimization_problem
        self.parallelization_backend = parallelization_backend

        super().__init__(
            n_var=optimization_problem.n_independent_variables,
            n_obj=optimization_problem.n_objectives,
            n_ieq_constr=optimization_problem.n_nonlinear_constraints,
            xl=optimization_problem.lower_bounds_independent_transformed,
            xu=optimization_problem.upper_bounds_independent_transformed,
            **kwargs,
        )

    def _evaluate(self, X: npt.ArrayLike, out: dict, *args: Any, **kwargs: Any) -> None:
        opt = self.optimization_problem
        if opt.n_objectives > 0:
            F = opt.evaluate_objectives(
                X,
                untransform=True,
                get_dependent_values=True,
                ensure_minimization=True,
                parallelization_backend=self.parallelization_backend,
            )
            out["F"] = np.array(F)

        if opt.n_nonlinear_constraints > 0:
            G = opt.evaluate_nonlinear_constraints(
                X,
                untransform=True,
                get_dependent_values=True,
                parallelization_backend=self.parallelization_backend,
            )
            CV = opt.evaluate_nonlinear_constraints_violation(
                X,
                untransform=True,
                get_dependent_values=True,
                parallelization_backend=self.parallelization_backend,
            )
            out["G"] = np.array(CV)

            out["CADET_G"] = G
            out["CADET_CV"] = CV


class RepairIndividuals(Repair):
    """Class to repair individuals."""

    def __init__(
        self,
        optimizer: OptimizerBase,
        optimization_problem: OptimizationProblem,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize repair individual object."""
        self.optimizer = optimizer
        self.optimization_problem = optimization_problem
        super().__init__(*args, **kwargs)

    def _do(self, problem: OptimizationProblem, X: npt.ArrayLike, **kwargs: Any) -> npt.ArrayLike:
        # Check if linear (equality) constraints are met
        X_new = None
        for i, ind in enumerate(X):
            if not self.optimization_problem.check_individual(
                ind,
                untransform=True,
                get_dependent_values=True,
                cv_bounds_tol=self.optimizer.cv_bounds_tol,
                cv_lincon_tol=self.optimizer.cv_lincon_tol,
                cv_lineqcon_tol=self.optimizer.cv_lineqcon_tol,
                check_nonlinear_constraints=False,
            ):
                if X_new is None:
                    X_new = self.optimization_problem.create_initial_values(
                        len(X), include_dependent_variables=False
                    )
                x_new = X_new[i, :]
                X[i, :] = self.optimization_problem.transform(x_new)

        return X



"""
===============================================
Optimization (:mod:`CADETProcess.optimization`)
===============================================

.. currentmodule:: CADETProcess.optimization

The ``optimization`` module provides functionality for minimizing (or maximizing)
objective functions, possibly subject to constraints. It includes interfaces to several
optimization suites, notably, ``scipy.optimize`` and ``pymoo``.

OptimizationProblem
===================
.. autosummary::
   :toctree: generated/

   OptimizationProblem


Optimizer
=========

Base
----

.. autosummary::
   :toctree: generated/

   OptimizerBase

Scipy
-----

.. autosummary::
   :toctree: generated/

   TrustConstr
   COBYLA
   NelderMead
   SLSQP

Pymoo
-----

.. autosummary::
   :toctree: generated/

   NSGA2
   U_NSGA3

Ax
--

.. autosummary::
   :toctree: generated/

   BotorchModular
   GPEI
   NEHVI

Population
==========
.. autosummary::
   :toctree: generated/

   Individual
   Population


Results
=======
.. autosummary::
   :toctree: generated/

   OptimizationResults


Cache
=====
.. autosummary::
   :toctree: generated/

   ResultsCache

ParallelizationBackend
======================
.. autosummary::
   :toctree: generated/

   ParallelizationBackendBase
   SequentialBackend
   Joblib
   Pathos

"""

from .individual import *
from .population import *
from .cache import *
from .results import *
from .optimizationProblem import *
from .parallelizationBackend import *
from .optimizer import *
from .scipyAdapter import COBYLA, TrustConstr, NelderMead, SLSQP
from .pymooAdapter import NSGA2, U_NSGA3

import importlib

try:
    from .axAdapater import BotorchModular, GPEI, NEHVI, qNParEGO

    ax_imported = True
except ImportError:
    ax_imported = False


def __getattr__(name):
    if name in ("BotorchModular", "GPEI", "NEHVI", "qNParEGO"):
        if ax_imported:
            module = importlib.import_module("axAdapter", package=__name__)
            return getattr(module, name)
        else:
            raise ImportError(
                "The AxInterface class could not be imported. "
                "This may be because the 'ax' package, which is an optional dependency, is not installed. "
                "To install it, run 'pip install CADET-Process[ax]'"
            )
    raise AttributeError(f"module {__name__} has no attribute {name}")



"""
==========================================================
Simulation Results (:mod:`CADETProcess.simulationResults`)
==========================================================

.. currentmodule:: CADETProcess.simulationResults

This module provides a class for storing simulation results.


.. autosummary::
    :toctree: generated/

    SimulationResults

"""  # noqa

from __future__ import annotations

import copy

import numpy as np
from addict import Dict

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import (
    Dictionary,
    List,
    String,
    Structure,
    UnsignedFloat,
    UnsignedInteger,
)
from CADETProcess.processModel import ComponentSystem, Process, UnitBaseClass
from CADETProcess.solution import SolutionBase

__all__ = ["SimulationResults"]


class SimulationResults(Structure):
    """
    Class for storing simulation results including the solver configuration.

    Attributes
    ----------
    solver_name : str
        Name of the solver used to simulate the process
    solver_parameters : dict
        Dictionary with parameters used by the solver
    exit_flag : int
        Information about the solver termination.
    exit_message : str
        Additional information about the solver status
    time_elapsed : float
        Execution time of simulation.
    process: Process
        Simulated Process.
    solution : dict
        Solution objects for all cycles of all Unit Operations.
    solution_cycles : dict
        Solution objects  for individual cycles of all Unit Operations.
    sensitivity : dict
        Solution objects for all sensitivities of all cycles of all Unit Operations.
    sensitivity_cycles : dict
        Solution objects for all sensitivities of individual cycles of all Unit Operations.
    system_state : dict
        Final state and state_derivative of the system.
    chromatograms : List of chromatogram
        Solution of the final cycle of the product outlets.
    n_cycles : int
        Number of cycles that were simulated.

    Notes
    -----
        Ideally, the final state for each unit operation should be saved.
        However, CADET does currently provide this functionality.
    """

    solver_name = String()
    solver_parameters = Dictionary()
    exit_flag = UnsignedInteger()
    exit_message = String()
    time_elapsed = UnsignedFloat()
    solution_cycles = Dictionary()
    sensitivity_cycles = Dictionary()
    system_state = Dictionary()
    chromatograms = List()

    def __init__(
        self,
        solver_name: str,
        solver_parameters: dict,
        exit_flag: int,
        exit_message: str,
        time_elapsed: float,
        process: Process,
        solution_cycles: dict,
        sensitivity_cycles: dict,
        system_state: dict,
        chromatograms: list,
    ) -> None:
        """Initialize SimulationResults."""
        self.solver_name = solver_name
        self.solver_parameters = solver_parameters

        self.exit_flag = exit_flag
        self.exit_message = exit_message
        self.time_elapsed = time_elapsed

        self.process = process

        self.solution_cycles = solution_cycles
        self.sensitivity_cycles = sensitivity_cycles
        self.system_state = system_state
        self.chromatograms = chromatograms

        self._time_complete = None
        self._solution = None
        self._sensitivity = None

    def update(self, new_results: SimulationResults) -> None:
        """Update the simulation results with results from a new cycle."""
        if self.process.name != new_results.process.name:
            raise CADETProcessError("Process does not match")

        self.exit_flag = new_results.exit_flag
        self.exit_message = new_results.exit_message
        self.time_elapsed += new_results.time_elapsed

        self.system_state = new_results.system_state

        self.chromatograms = new_results.chromatograms
        for unit, solutions in self.solution_cycles.items():
            for sol in solutions:
                solution = new_results.solution_cycles[unit][sol]
                self.solution_cycles[unit][sol] += solution

        self._time_complete = None
        self._solution = None
        self._sensitivity = None

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: The component system used in the simulation."""
        solution = self.solution_cycles[self._first_unit][self._first_solution]
        return solution[0].component_system

    @property
    def solution(self) -> Dict:
        """Construct complete solution from individual cyles."""
        if self._solution is not None:
            return self._solution

        time_complete = self.time_complete

        solution = Dict()
        for unit, solutions in self.solution_cycles.items():
            for sol, ports_cycles in solutions.items():
                if isinstance(ports_cycles, Dict):
                    ports = ports_cycles
                    for port, cycles in ports.items():
                        solution[unit][sol][port] = copy.deepcopy(cycles[0])
                        solution_complete = cycles[0].solution
                        if solution_complete.ndim > 1:
                            for i in range(1, self.n_cycles):
                                solution_complete = np.vstack((
                                    solution_complete, cycles[i].solution[1:]
                                ))
                        else:
                            for i in range(1, self.n_cycles):
                                solution_complete = np.hstack((
                                    solution_complete, cycles[i].solution[1:]
                                ))

                        solution[unit][sol][port].time = time_complete
                        solution[unit][sol][port].solution = solution_complete
                        solution[unit][sol][port].update_solution()
                else:
                    cycles = ports_cycles
                    solution[unit][sol] = copy.deepcopy(cycles[0])
                    solution_complete = cycles[0].solution
                    if solution_complete.ndim > 1:
                        for i in range(1, self.n_cycles):
                            solution_complete = np.vstack((
                                solution_complete, cycles[i].solution[1:]
                            ))
                    else:
                        for i in range(1, self.n_cycles):
                            solution_complete = np.hstack((
                                solution_complete, cycles[i].solution[1:]
                            ))

                    solution[unit][sol].time = time_complete
                    solution[unit][sol].solution = solution_complete
                    solution[unit][sol].update_solution()

        self._solution = solution

        return solution

    @property
    def sensitivity(self) -> Dict:
        """Construct complete sensitivity from individual cyles."""
        if self._sensitivity is not None:
            return self._sensitivity

        time_complete = self.time_complete

        sensitivity = Dict()
        for sens_name, sensitivities in self.sensitivity_cycles.items():
            for unit, sensitivities in sensitivities.items():
                for flow, ports_cycles in sensitivities.items():
                    if isinstance(ports_cycles, Dict):
                        ports = ports_cycles
                        for port, cycles in ports.items():
                            sensitivity[sens_name][unit][flow][port] = copy.deepcopy(
                                cycles[0]
                            )
                            sensitivity_complete = cycles[0].solution
                            for i in range(1, self.n_cycles):
                                sensitivity_complete = np.vstack((
                                    sensitivity_complete, cycles[i].solution[1:]
                                ))
                            sensitivity[sens_name][unit][flow][port].time = time_complete
                            sensitivity[sens_name][unit][flow][port].solution = sensitivity_complete
                            sensitivity[sens_name][unit][flow][port].update_solution()

                    else:
                        cycles = ports_cycles
                        sensitivity[sens_name][unit][flow] = copy.deepcopy(cycles[0])
                        sensitivity_complete = cycles[0].solution
                        for i in range(1, self.n_cycles):
                            sensitivity_complete = np.vstack((
                                sensitivity_complete, cycles[i].solution[1:]
                            ))
                        sensitivity[sens_name][unit][flow].time = time_complete
                        sensitivity[sens_name][unit][flow].solution = sensitivity_complete
                        sensitivity[sens_name][unit][flow].update_solution()

        self._sensitivity = sensitivity

        return sensitivity

    @property
    def n_cycles(self) -> int:
        """int: Number of simulated cycles."""
        return len(self.solution_cycles[self._first_unit][self._first_solution])

    @property
    def _first_unit(self) -> UnitBaseClass:
        return next(iter(self.solution_cycles))

    @property
    def _first_solution(self) -> SolutionBase:
        return next(iter(self.solution_cycles[self._first_unit]))

    @property
    def time_cycle(self) -> np.ndarray:
        """np.array: Solution times vector."""
        return self.solution_cycles[self._first_unit][self._first_solution][0].time

    @property
    def time_complete(self) -> np.ndarray:
        """np.ndarray: Solution times vector for all cycles."""
        if self._time_complete is not None:
            return self._time_complete

        time_complete = self.time_cycle
        for i in range(1, self.n_cycles):
            time_complete = np.hstack((
                time_complete, self.time_cycle[1:] + i * self.process.cycle_time
            ))

        self._time_complete = time_complete

        return time_complete



from CADETProcess.dataStructure import Bool, Structure


class BaseMixin(Structure):
    """
    Recorder for last state and coordinates.

    Attributes
    ----------
    write_solution_last_unit : bool, optional
        If True, write final state of the unit operation. The default is False.
    write_coordinates : bool, optional
        If True, write coordinates of the unit operation. The default is True.
    """

    write_solution_last_unit = Bool(default=False)
    write_coordinates = Bool(default=True)

    _parameters = [
        "write_coordinates",
        "write_solution_last_unit",
    ]


class IOMixin(Structure):
    """
    Recorder for inlet and outlet streams.

    Attributes
    ----------
    write_solution_inlet : bool, optional
        If True, write solution of the unit operation inlet. The default is True.
    write_soldot_inlet : bool, optional
        If True, write derivative of the unit operation inlet. The default is False.
    write_sens_inlet : bool, optional
        If True, write sensitivities of the unit operation inlet. The default is True.
    write_sensdot_inlet : bool, optional
        If True, write sensitivities derivatives of the unit operation inlet.
        The default is False.
    write_solution_outlet : bool, optional
        If True, write solution of the unit operation outlet. The default is True.
    write_soldot_outlet : bool, optional
        If True, write derivative of the unit operation outlet. The default is False.
    write_sens_outlet : bool, optional
        If True, write sensitivities of the unit operation outlet. The default is True.
    write_sensdot_outlet : bool, optional
        If True, write sensitivities derivatives of the unit operation outlet.
        The default is False.
    """

    write_solution_inlet = Bool(default=True)
    write_soldot_inlet = Bool(default=False)
    write_sens_inlet = Bool(default=True)
    write_sensdot_inlet = Bool(default=False)

    write_solution_outlet = Bool(default=True)
    write_soldot_outlet = Bool(default=False)
    write_sens_outlet = Bool(default=True)
    write_sensdot_outlet = Bool(default=False)

    _parameters = [
        "write_solution_inlet",
        "write_soldot_inlet",
        "write_sens_inlet",
        "write_sensdot_inlet",
        "write_solution_outlet",
        "write_soldot_outlet",
        "write_sens_outlet",
        "write_sensdot_outlet",
    ]


class BulkMixin(Structure):
    """
    Recorder for bulk solution.

    Attributes
    ----------
    write_solution_bulk : bool, optional
        If True, write solution of the bulk. The default is False.
    write_soldot_bulk : bool, optional
        If True, write derivative of the bulk. The default is False.
    write_sens_bulk : bool, optional
        If True, write sensitivities of the bulk. The default is False.
    write_sensdot_bulk : bool, optional
        If True, write sensitivities derivatives of the bulk. The default is False.
    """

    write_solution_bulk = Bool(default=False)
    write_soldot_bulk = Bool(default=False)
    write_sens_bulk = Bool(default=False)
    write_sensdot_bulk = Bool(default=False)

    _parameters = [
        "write_solution_bulk",
        "write_soldot_bulk",
        "write_sens_bulk",
        "write_sensdot_bulk",
    ]


class ParticleMixin(Structure):
    """
    Recorder for particle liquid phase solution.

    Attributes
    ----------
    write_solution_particle : bool, optional
        If True, write the solution of the particle liquid phase. The default is False.
    write_soldot_particle : bool, optional
        If True, write the derivative of the particle liquid phase.
        The default is False.
    write_sens_particle : bool, optional
        If True, write the sensitivities of the particle liquid phase.
        The default is False.
    write_sensdot_particle : bool, optional
        If True, write the sensitivities derivatives of the particle liquid phase.
        The default is False.
    """

    write_solution_particle = Bool(default=False)
    write_soldot_particle = Bool(default=False)
    write_sens_particle = Bool(default=False)
    write_sensdot_particle = Bool(default=False)

    _parameters = [
        "write_solution_particle",
        "write_soldot_particle",
        "write_sens_particle",
        "write_sensdot_particle",
    ]


class SolidMixin(Structure):
    """
    Recorder for particle solid phase solution.

    Attributes
    ----------
    write_solution_solid : bool, optional
        If True, write the solution of the particle solid phase. The default is False.
    write_soldot_solid : bool, optional
        If True, write the derivative of the particle solid phase. The default is False.
    write_sens_solid : bool, optional
        If True, write the sensitivities of the particle solid phase.
        The default is False.
    write_sensdot_solid : bool, optional
        If True, write the sensitivities derivatives of the particle solid phase.
        The default is False.
    """

    write_solution_solid = Bool(default=False)
    write_soldot_solid = Bool(default=False)
    write_sens_solid = Bool(default=False)
    write_sensdot_solid = Bool(default=False)

    _parameters = [
        "write_solution_solid",
        "write_soldot_solid",
        "write_sens_solid",
        "write_sensdot_solid",
    ]


class FluxMixin(Structure):
    """
    Recorder for flux solution.

    Attributes
    ----------
    write_solution_flux : bool, optional
        If True, write the solution of the flux. The default is False.
    write_soldot_flux : bool, optional
        If True, write the derivative of the flux. The default is False.
    write_sens_flux : bool, optional
        If True, write the sensitivities of the flux. The default is False.
    write_sensdot_flux : bool, optional
        If True, write the sensitivities derivatives of the flux. The default is False.
    """

    write_solution_flux = Bool(default=False)
    write_soldot_flux = Bool(default=False)
    write_sensdot_flux = Bool(default=False)
    write_sens_flux = Bool(default=False)

    _parameters = [
        "write_solution_flux",
        "write_soldot_flux",
        "write_sens_flux",
        "write_sensdot_flux",
    ]


class VolumeMixin(Structure):
    """
    Recorder for unit volume solution.

    Attributes
    ----------
    write_solution_volume : bool, optional
        If True, write the solution of the unit volume. The default is True.
    write_soldot_volume : bool, optional
        If True, write the derivative of the unit volume. The default is False.
    write_sens_volume : bool, optional
        If True, write the sensitivities of the unit volume. The default is False.
    write_sensdot_volume : bool, optional
        If True, write the sensitivities derivatives of the unit volume.
        The default is False.
    """

    write_solution_volume = Bool(default=True)
    write_soldot_volume = Bool(default=False)
    write_sens_volume = Bool(default=False)
    write_sensdot_volume = Bool(default=False)

    _parameters = [
        "write_solution_volume",
        "write_soldot_volume",
        "write_sens_volume",
        "write_sensdot_volume",
    ]


class SolutionRecorderBase:
    """Base class for solution recorders."""

    pass


class IORecorder(SolutionRecorderBase, BaseMixin, IOMixin):
    """
    Recorder for inlets and outlets.

    See Also
    --------
    BaseMixin
    IOMixin
    CADETProcess.processModel.Inlet
    CADETProcess.processModel.Outlet
    """

    pass


class TubularReactorRecorder(SolutionRecorderBase, BaseMixin, IOMixin, BulkMixin):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    SolutionRecorderBulk
    CADETProcess.processModel.TubularReactor
    """

    pass


class LRMRecorder(SolutionRecorderBase, BaseMixin, IOMixin, BulkMixin, SolidMixin):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    BulkMixin
    SolidMixin
    CADETProcess.processModel.LumpedRateModelWithoutPores
    """

    pass


class LRMPRecorder(
    SolutionRecorderBase,
    BaseMixin,
    IOMixin,
    BulkMixin,
    FluxMixin,
    ParticleMixin,
    SolidMixin,
):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    BulkMixin
    FluxMixin
    ParticleMixin
    SolidMixin
    CADETProcess.processModel.LumpedRateModelWithPores
    """

    pass


class GRMRecorder(
    SolutionRecorderBase,
    BaseMixin,
    IOMixin,
    BulkMixin,
    FluxMixin,
    ParticleMixin,
    SolidMixin,
):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    BulkMixin
    FluxMixin
    ParticleMixin
    SolidMixin
    CADETProcess.processModel.GeneralRateModel
    """

    pass


class CSTRRecorder(
    SolutionRecorderBase, BaseMixin, IOMixin, BulkMixin, SolidMixin, VolumeMixin
):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    BulkMixin
    SolidMixin
    VolumeMixin
    CADETProcess.processModel.Cstr
    """

    pass


class MCTRecorder(SolutionRecorderBase, BaseMixin, IOMixin, BulkMixin):
    """
    Recorder for TubularReactor.

    See Also
    --------
    BaseMixin
    IOMixin
    BulkMixin
    CADETProcess.processModel.MCT
    """

    pass



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     name: python3
# ---

# %% [markdown]
# (ssr_process)=
# # Steady-State Recycling Process
# In addition to the recycled fraction, fresh feed can also be injected in each cycle, resulting in the formation of a cyclic steady-state.
# This process, called closed-loop steady-state recycling (CL-SSR), can achieve higher productivity compared to CLR.
# However, due to additional dispersion in the system periphery, maintaining the separation of components generated during the passage of the column is difficult to realize.
# Hence, determining the optimal time at which to add new feed is therefore complex.
# To overcome this problem, a tank can be inserted in which the recycling fraction and new feed are mixed.
# The recycling fraction and new feed are then injected together in a process called mixed-recycle steady-state recycling (MR-SSR).
# A schematic flow diagram of the MR-SSR process is shown below.
#
# ```{figure} ./figures/mrssr_flow_sheet.svg
# :name: mrssr_flow_sheet
#
# Flow sheet for mixed-recycle steady-state recycling process.
# ```
#
# For this demonstration, consider a two-component system with a Langmuir isotherm.
#
# ## Component System

# %%
from CADETProcess.processModel import ComponentSystem

component_system = ComponentSystem(['A', 'B'])

# %% [markdown]
# ## Binding Model

# %%
from CADETProcess.processModel import Langmuir

binding_model = Langmuir(component_system, name='langmuir')
binding_model.is_kinetic = False
binding_model.adsorption_rate = [0.02, 0.03]
binding_model.desorption_rate = [1, 1]
binding_model.capacity = [100, 100]

# %% [markdown]
# ## Unit Operations

# %%
from CADETProcess.processModel import (
    Inlet, Cstr, LumpedRateModelWithoutPores, Outlet
)
feed = Inlet(component_system, name='feed')
feed.c = [10, 10]

eluent = Inlet(component_system, name='eluent')
eluent.c = [0, 0]

tank = Cstr(component_system, name='tank')
tank.V = 0.001
tank.c = [10, 10]

column = LumpedRateModelWithoutPores(component_system, name='column')
column.binding_model = binding_model

column.length = 0.6
column.diameter = 0.024
column.axial_dispersion = 4.7e-7
column.total_porosity = 0.7

column.solution_recorder.write_solution_bulk = True

outlet = Outlet(component_system, name='outlet')

# %% [markdown]
# ## Flow Sheet

# %%
from CADETProcess.processModel import FlowSheet

flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(feed, feed_inlet=True)
flow_sheet.add_unit(eluent, eluent_inlet=True)
flow_sheet.add_unit(tank)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet, product_outlet=True)

flow_sheet.add_connection(feed, tank)
flow_sheet.add_connection(tank, column)
flow_sheet.add_connection(eluent, column)
flow_sheet.add_connection(column, tank)
flow_sheet.add_connection(column, outlet)

# %% [markdown]
# ## Process
# To realize the recycling, the {attr}`~CADETProcess.processModel.FlowSheet.output_state` of the column needs to be modified.
# To reduce the number of event times that need to be specified, event dependencies are specified which enforce that always either feed or eluent are being pumped through the column.
#
# ```{figure} ./figures/mrssr_events.svg
# :name: mrssr_events
#
# Events for mixed-recycle steady-state recycling process with event dependencies.
# ```

# %%
from CADETProcess.processModel import Process
process = Process(flow_sheet, 'mrssr')

Q = 60 / (60 * 1e6)

# Create Events and Durations
process.add_event('feed_on', 'flow_sheet.feed.flow_rate', Q)
process.add_event('feed_off', 'flow_sheet.feed.flow_rate', 0.0)
process.add_duration('feed_duration')

process.add_event('inject_on', 'flow_sheet.tank.flow_rate', Q)
process.add_event('inject_off', 'flow_sheet.tank.flow_rate', 0.0)

process.add_event('eluent_on', 'flow_sheet.eluent.flow_rate', Q)
process.add_event_dependency('eluent_on', ['inject_off'])
process.add_event('eluent_off', 'flow_sheet.eluent.flow_rate', 0.0)
process.add_event_dependency('eluent_off', ['inject_on'])

process.add_event('recycle_on', 'flow_sheet.output_states.column', 0)
process.add_event('recycle_off', 'flow_sheet.output_states.column', 1)

process.add_event_dependency('feed_on', ['inject_off'])
process.add_event_dependency('feed_off', ['feed_on', 'feed_duration'], [1, 1])
process.add_event_dependency(
    'inject_off', ['inject_on', 'feed_duration', 'recycle_off', 'recycle_on'],
    [1, 1, 1, -1]
)

# %% [markdown]
# Now, the cycle time is set to $10~min$ and the `feed_duration` to $1~min$ and the recycling times are specified.

# %%
process.cycle_time = 600
process.feed_duration.time = 60
process.recycle_on.time = 360
process.recycle_off.time = 420

# %% [markdown]
# ## Simulate Process
# Since the process shows a startup behavior before reaching steady state, multiple cycles need to be simulated.
# For this purpose, a {class}`~CADETProcess.stationarity.StationarityEvaluator` is used (see {ref}`stationarity_guide`).

# %%
if __name__ == '__main__':
    from CADETProcess.simulator import Cadet
    process_simulator = Cadet()
    process_simulator.evaluate_stationarity = True

    simulation_results = process_simulator.simulate(process)
    simulation_results.solution.column.outlet.plot()
    simulation_results.solution_cycles.column.outlet[-1].plot()

# %% [markdown]
# ## Optimize Fractionation Times
# To evaluate the process performance, fractionation times can determined to maximize yield and ensure purity constraints.

# %%
if __name__ == '__main__':
    from CADETProcess.fractionation import FractionationOptimizer
    fractionation_optimizer = FractionationOptimizer()

    fractionator = fractionation_optimizer.optimize_fractionation(
        simulation_results, purity_required=[0.95, 0.95]
    )

    print(fractionator.performance)
    _ = fractionator.plot_fraction_signal()



import unittest

import numpy as np
from CADETProcess import equilibria
from CADETProcess.processModel import ComponentSystem, MassActionLaw

enable_plot = False


class TestBufferCapacity(unittest.TestCase):
    def setUp(self):
        self.components_simple = ComponentSystem(2, charges=[1, 2])

        # Ammonia
        self.components_ammonia = ComponentSystem()
        self.components_ammonia.add_component(
            "Ammonia", species=["NH4+", "NH3"], charge=[1, 0]
        )
        self.components_ammonia.add_component("H+", charge=1)

        self.reaction_ammonia = MassActionLaw(self.components_ammonia)
        self.reaction_ammonia.add_reaction(
            [0, 1, 2], [-1, 1, 1], 10 ** (-9.2) * 1e3, is_kinetic=False
        )

        # Ammonia, component order switched
        self.components_ammonia_switched = ComponentSystem()
        self.components_ammonia_switched.add_component("H+", charge=1)
        self.components_ammonia_switched.add_component(
            "Ammonia", species=["NH4+", "NH3"], charge=[1, 0]
        )

        self.reaction_ammonia_switched = MassActionLaw(self.components_ammonia_switched)
        self.reaction_ammonia_switched.add_reaction(
            [1, 2, 0], [-1, 1, 1], 10 ** (-9.2) * 1e3, is_kinetic=False
        )

        # Lysine
        self.components_lys = ComponentSystem()
        self.components_lys.add_component(
            "Lysine", species=["Lys2+", "Lys+", "Lys", "Lys-"], charge=[2, 1, 0, -1]
        )
        self.components_lys.add_component("H+", charge=1)
        self.reaction_lys = MassActionLaw(self.components_lys)
        self.reaction_lys.add_reaction(
            [0, 1, -1], [-1, 1, 1], 10 ** (-2.20) * 1e3, is_kinetic=False
        )
        self.reaction_lys.add_reaction(
            [1, 2, -1], [-1, 1, 1], 10 ** (-8.90) * 1e3, is_kinetic=False
        )
        self.reaction_lys.add_reaction(
            [2, 3, -1], [-1, 1, 1], 10 ** (-10.28) * 1e3, is_kinetic=False
        )

        # Lysine, component order switched
        self.components_lys_switched = ComponentSystem()
        self.components_lys_switched.add_component("H+", charge=1)
        self.components_lys_switched.add_component(
            "Lysine", species=["Lys2+", "Lys+", "Lys", "Lys-"], charge=[2, 1, 0, -1]
        )

        self.reaction_lys_switched = MassActionLaw(self.components_lys_switched)
        self.reaction_lys_switched.add_reaction(
            [1, 2, 0], [-1, 1, 1], 10 ** (-2.20) * 1e3, is_kinetic=False
        )
        self.reaction_lys_switched.add_reaction(
            [2, 3, 0], [-1, 1, 1], 10 ** (-8.90) * 1e3, is_kinetic=False
        )
        self.reaction_lys_switched.add_reaction(
            [3, 4, 0], [-1, 1, 1], 10 ** (-10.28) * 1e3, is_kinetic=False
        )

        # Ammonia and Lysine
        self.components_ammonia_lys = ComponentSystem()
        self.components_ammonia_lys.add_component(
            "Ammonia", species=["NH4+", "NH3"], charge=[1, 0]
        )
        self.components_ammonia_lys.add_component(
            "Lysine", species=["Lys2+", "Lys+", "Lys", "Lys-"], charge=[2, 1, 0, -1]
        )
        self.components_ammonia_lys.add_component("H+", charge=1)
        self.reaction_ammonia_lys = MassActionLaw(self.components_ammonia_lys)
        self.reaction_ammonia_lys.add_reaction(
            [0, 1, -1], [-1, 1, 1], 10 ** (-9.2) * 1e3, is_kinetic=False
        )
        self.reaction_ammonia_lys.add_reaction(
            [2, 3, -1], [-1, 1, 1], 10 ** (-2.20) * 1e3, is_kinetic=False
        )
        self.reaction_ammonia_lys.add_reaction(
            [3, 4, -1], [-1, 1, 1], 10 ** (-8.90) * 1e3, is_kinetic=False
        )
        self.reaction_ammonia_lys.add_reaction(
            [4, 5, -1], [-1, 1, 1], 10 ** (-10.28) * 1e3, is_kinetic=False
        )

    def test_ionic_strength(self):
        c = [1, 1]
        i_excpected = 2.5
        i = equilibria.ionic_strength(self.components_simple, c)
        np.testing.assert_almost_equal(i, i_excpected)

        c = [1, 2]
        i_excpected = 4.5
        i = equilibria.ionic_strength(self.components_simple, c)
        np.testing.assert_almost_equal(i, i_excpected)

        c = [1, 0, 1]
        i_excpected = 1
        i = equilibria.ionic_strength(self.components_ammonia, c)
        np.testing.assert_almost_equal(i, i_excpected)

        # Uncharged species should have no effect
        c = [1, 1, 1]
        i_excpected = 1
        i = equilibria.ionic_strength(self.components_ammonia, c)
        np.testing.assert_almost_equal(i, i_excpected)

        c = [2, 0, 1]
        i_excpected = 1.5
        i = equilibria.ionic_strength(self.components_ammonia, c)
        np.testing.assert_almost_equal(i, i_excpected)

    def test_buffer_capacity(self):
        buffer = [0, 1000, 0]

        pH = 0
        b_expected = np.array([1.4528329738818527e-06, 2302.585092994069])
        b = equilibria.buffer_capacity(self.reaction_ammonia, buffer, pH)
        np.testing.assert_almost_equal(b_expected, b)

        pH = 7
        b_expected = np.array([14.346715296419855, 0.0004605170185988092])
        b = equilibria.buffer_capacity(self.reaction_ammonia, buffer, pH)
        np.testing.assert_almost_equal(b_expected, b)

        pH = 10
        b_expected = np.array([271.9140324242618, 0.23025873955791384])
        b = equilibria.buffer_capacity(self.reaction_ammonia, buffer, pH)
        np.testing.assert_almost_equal(b_expected, b)

        pH = 14
        b_expected = np.array([0.03649235765053616, 2302.585092994069])
        b = equilibria.buffer_capacity(self.reaction_ammonia, buffer, pH)
        np.testing.assert_almost_equal(b_expected, b)

        pH = 0
        b_expected = np.array([1.4528329738818527e-06, 2302.585092994069])
        b = equilibria.buffer_capacity(self.reaction_ammonia_switched, buffer, pH)
        np.testing.assert_almost_equal(b_expected, b)

    def test_charge_distribution(self):
        pH = 0
        eta_expected = np.array([0.9999999993690426, 6.30957344082087e-10])
        eta = equilibria.charge_distribution(self.reaction_ammonia, pH)
        np.testing.assert_almost_equal(eta_expected, eta)

        pH = 7
        eta_expected = np.array([0.9937299876585661, 0.006270012341433852])
        eta = equilibria.charge_distribution(self.reaction_ammonia, pH)
        np.testing.assert_almost_equal(eta_expected, eta)

        pH = 10
        eta_expected = np.array([0.13680688860320983, 0.8631931113967902])
        eta = equilibria.charge_distribution(self.reaction_ammonia, pH)
        np.testing.assert_almost_equal(eta_expected, eta)

        pH = 14
        eta_expected = np.array([1.5848680739948967e-05, 0.99998415131926])
        eta = equilibria.charge_distribution(self.reaction_ammonia, pH)
        np.testing.assert_almost_equal(eta_expected, eta)

        pH = 0
        eta_expected = np.array([0.9999999993690426, 6.30957344082087e-10])
        eta = equilibria.charge_distribution(self.reaction_ammonia_switched, pH)
        np.testing.assert_almost_equal(eta_expected, eta)

    def test_plot(self):
        if enable_plot:
            _ = equilibria.plot_charge_distribution(self.reaction_ammonia)
            _ = equilibria.plot_charge_distribution(self.reaction_ammonia_switched)
            _ = equilibria.plot_charge_distribution(
                self.reaction_lys_switched, plot_cumulative=True
            )
            _ = equilibria.plot_charge_distribution(
                self.reaction_ammonia_switched, plot_cumulative=True
            )


if __name__ == "__main__":
    enable_plot = True
    unittest.main()



"""
=========================================
Simulator (:mod:`CADETProcess.simulator`)
=========================================

.. currentmodule:: CADETProcess.simulator

The ``simulator`` module provides functionality for simulating a ``Process``.

Simulator
=========

.. autosummary::
   :toctree: generated/

   SimulatorBase

CADET
-----

.. autosummary::
   :toctree: generated/

   Cadet

Futher settings:

.. autosummary::
   :toctree: generated/

   ModelSolverParameters
   UnitParameters
   AdsorptionParameters
   ReactionParameters
   SolverParameters
   SolverTimeIntegratorParameters
   ReturnParameters
   SensitivityParameters


SimulationResults
=================

After simulation, the ``Simulator`` returns a ``SimulationResults`` object.

.. currentmodule:: CADETProcess.simulationResults

.. autosummary::
   :toctree: generated/

   SimulationResults

"""

from .simulator import *
from .cadetAdapter import *



import shutil
import unittest

from CADETProcess import settings
from CADETProcess.optimization import U_NSGA3

from tests.test_optimization_problem import setup_optimization_problem


class Test_OptimizationProblemSimple(unittest.TestCase):
    def tearDown(self):
        shutil.rmtree("./results_simple", ignore_errors=True)
        settings.working_directory = None

    def test_restart_from_checkpoint(self):
        class Callback:
            def __init__(self, n_calls=0, kill=True):
                self.n_calls = n_calls
                self.kill = kill

            def __call__(self, results):
                if self.kill and self.n_calls == 2:
                    raise Exception("Kill this!")
                self.n_calls += 1

        callback = Callback()
        optimization_problem = setup_optimization_problem()
        optimization_problem.add_callback(callback)

        optimizer = U_NSGA3()
        optimizer.n_max_gen = 5

        try:
            opt_results = optimizer.optimize(
                optimization_problem,
                save_results=True,
                use_checkpoint=False,
            )
        except Exception:
            pass

        callback.kill = False

        optimization_problem = setup_optimization_problem()
        optimization_problem.add_callback(callback)

        opt_results = optimizer.optimize(
            optimization_problem,
            save_results=True,
            use_checkpoint=True,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from CADETProcess.modelBuilder import CarouselBuilder, ParallelZone, SerialZone
from CADETProcess.processModel import (
    ComponentSystem,
    Inlet,
    Linear,
    LumpedRateModelWithoutPores,
    Outlet,
)
from CADETProcess.simulator import Cadet


class Test_Carousel(unittest.TestCase):
    def setUp(self):
        self.component_system = ComponentSystem(2)

        self.binding_model = Linear(self.component_system)
        self.binding_model.adsorption_rate = [6, 8]
        self.binding_model.desorption_rate = [1, 1]

        self.column = LumpedRateModelWithoutPores(
            self.component_system, name="master_column"
        )
        self.column.length = 0.6
        self.column.diameter = 0.024
        self.column.axial_dispersion = 4.7e-7
        self.column.total_porosity = 0.7

        self.column.binding_model = self.binding_model

    def create_serial(self):
        source = Inlet(self.component_system, name="source")
        source.c = [10, 10]
        source.flow_rate = 2e-7

        sink = Outlet(self.component_system, name="sink")

        serial_zone = SerialZone(self.component_system, "serial", 2, flow_direction=1)

        builder = CarouselBuilder(self.component_system, "serial")
        builder.column = self.column

        builder.add_unit(source)
        builder.add_unit(sink)
        builder.add_unit(serial_zone)

        builder.add_connection(source, serial_zone)
        builder.add_connection(serial_zone, sink)

        builder.switch_time = 300

        return builder

    def create_parallel(self):
        source = Inlet(self.component_system, name="source")
        source.c = [10, 10]
        source.flow_rate = 2e-7

        sink = Outlet(self.component_system, name="sink")

        parallel_zone = ParallelZone(
            self.component_system, "parallel", 2, flow_direction=1
        )

        builder = CarouselBuilder(self.component_system, "parallel")
        builder.column = self.column

        builder.add_unit(source)
        builder.add_unit(sink)
        builder.add_unit(parallel_zone)

        builder.add_connection(source, parallel_zone)
        builder.add_connection(parallel_zone, sink)

        builder.switch_time = 300

        return builder

    def create_smb(self):
        feed = Inlet(self.component_system, name="feed")
        feed.c = [10, 10]
        feed.flow_rate = 2e-7

        eluent = Inlet(self.component_system, name="eluent")
        eluent.c = [0, 0]
        eluent.flow_rate = 6e-7

        raffinate = Outlet(self.component_system, name="raffinate")
        extract = Outlet(self.component_system, name="extract")

        zone_I = SerialZone(self.component_system, "zone_I", 1)
        zone_II = SerialZone(self.component_system, "zone_II", 1)
        zone_III = SerialZone(self.component_system, "zone_III", 1)
        zone_IV = SerialZone(self.component_system, "zone_IV", 1)

        builder = CarouselBuilder(self.component_system, "smb")
        builder.column = self.column
        builder.add_unit(feed)
        builder.add_unit(eluent)

        builder.add_unit(raffinate)
        builder.add_unit(extract)

        builder.add_unit(zone_I)
        builder.add_unit(zone_II)
        builder.add_unit(zone_III)
        builder.add_unit(zone_IV)

        builder.add_connection(eluent, zone_I)

        builder.add_connection(zone_I, extract)
        builder.add_connection(zone_I, zone_II)
        w_e = 0.15
        builder.set_output_state(zone_I, [w_e, 1 - w_e])

        builder.add_connection(zone_II, zone_III)

        builder.add_connection(feed, zone_III)

        builder.add_connection(zone_III, raffinate)
        builder.add_connection(zone_III, zone_IV)
        w_r = 0.15
        builder.set_output_state(zone_III, [w_r, 1 - w_r])

        builder.add_connection(zone_IV, zone_I)

        builder.switch_time = 300

        return builder

    def create_multi_zone(self):
        source_serial = Inlet(self.component_system, name="source_serial")
        source_serial.c = [10, 10]
        source_serial.flow_rate = 2e-7

        sink_serial = Outlet(self.component_system, name="sink_serial")

        serial_zone = SerialZone(self.component_system, "serial", 2, flow_direction=1)

        source_parallel = Inlet(self.component_system, name="source_parallel")
        source_parallel.c = [10, 10]
        source_parallel.flow_rate = 2e-7

        sink_parallel = Outlet(self.component_system, name="sink_parallel")

        parallel_zone = ParallelZone(
            self.component_system, "parallel", 2, flow_direction=-1
        )

        builder = CarouselBuilder(self.component_system, "multi_zone")
        builder.column = self.column
        builder.add_unit(source_serial)
        builder.add_unit(source_parallel)

        builder.add_unit(sink_serial)
        builder.add_unit(sink_parallel)

        builder.add_unit(serial_zone)
        builder.add_unit(parallel_zone)

        builder.add_connection(source_serial, serial_zone)
        builder.add_connection(serial_zone, sink_serial)
        builder.add_connection(serial_zone, parallel_zone)
        builder.set_output_state(serial_zone, [0.5, 0.5])

        builder.add_connection(source_parallel, parallel_zone)
        builder.add_connection(parallel_zone, sink_parallel)

        builder.switch_time = 300

        return builder

    def test_units(self):
        """Check if all units are added properly in the FlowSheet"""
        # Serial
        builder = self.create_serial()
        flow_sheet = builder.build_flow_sheet()

        units_expected = [
            "source",
            "sink",
            "serial_inlet",
            "serial_outlet",
            "column_0",
            "column_1",
        ]

        self.assertEqual(units_expected, flow_sheet.unit_names)

        # Parallel
        builder = self.create_parallel()
        flow_sheet = builder.build_flow_sheet()

        units_expected = [
            "source",
            "sink",
            "parallel_inlet",
            "parallel_outlet",
            "column_0",
            "column_1",
        ]

        self.assertEqual(units_expected, flow_sheet.unit_names)

        # SMB
        builder = self.create_smb()
        flow_sheet = builder.build_flow_sheet()

        units_expected = [
            "feed",
            "eluent",
            "raffinate",
            "extract",
            "zone_I_inlet",
            "zone_I_outlet",
            "column_0",
            "zone_II_inlet",
            "zone_II_outlet",
            "column_1",
            "zone_III_inlet",
            "zone_III_outlet",
            "column_2",
            "zone_IV_inlet",
            "zone_IV_outlet",
            "column_3",
        ]

        self.assertEqual(units_expected, flow_sheet.unit_names)

    def test_connections(self):
        """Check if all units are connected properly in the FlowSheet"""
        # Serial
        builder = self.create_serial()
        flow_sheet = builder.build_flow_sheet()

        self.assertTrue(flow_sheet.connection_exists("source", "serial_inlet"))

        self.assertTrue(flow_sheet.connection_exists("serial_inlet", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("serial_inlet", "column_1"))

        self.assertTrue(flow_sheet.connection_exists("column_0", "column_1"))
        self.assertTrue(flow_sheet.connection_exists("column_0", "serial_outlet"))

        self.assertTrue(flow_sheet.connection_exists("column_1", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("column_1", "serial_outlet"))

        # Parallel
        builder = self.create_parallel()
        flow_sheet = builder.build_flow_sheet()

        self.assertTrue(flow_sheet.connection_exists("source", "parallel_inlet"))

        self.assertTrue(flow_sheet.connection_exists("parallel_inlet", "column_0"))

        self.assertTrue(flow_sheet.connection_exists("parallel_inlet", "column_1"))

        self.assertTrue(flow_sheet.connection_exists("column_0", "parallel_outlet"))

        self.assertTrue(flow_sheet.connection_exists("column_1", "parallel_outlet"))

        # SMB
        builder = self.create_smb()
        flow_sheet = builder.build_flow_sheet()

        self.assertTrue(flow_sheet.connection_exists("eluent", "zone_I_inlet"))
        self.assertTrue(flow_sheet.connection_exists("feed", "zone_III_inlet"))

        self.assertTrue(flow_sheet.connection_exists("zone_I_outlet", "extract"))
        self.assertTrue(flow_sheet.connection_exists("zone_I_outlet", "zone_II_inlet"))

        self.assertTrue(
            flow_sheet.connection_exists("zone_II_outlet", "zone_III_inlet")
        )

        self.assertTrue(flow_sheet.connection_exists("zone_III_outlet", "raffinate"))
        self.assertTrue(
            flow_sheet.connection_exists("zone_III_outlet", "zone_IV_inlet")
        )

        self.assertTrue(flow_sheet.connection_exists("zone_IV_outlet", "zone_I_inlet"))

        self.assertTrue(flow_sheet.connection_exists("zone_I_inlet", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("zone_I_inlet", "column_1"))
        self.assertTrue(flow_sheet.connection_exists("zone_I_inlet", "column_2"))
        self.assertTrue(flow_sheet.connection_exists("zone_I_inlet", "column_3"))

        self.assertTrue(flow_sheet.connection_exists("zone_II_inlet", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("zone_II_inlet", "column_1"))
        self.assertTrue(flow_sheet.connection_exists("zone_II_inlet", "column_2"))
        self.assertTrue(flow_sheet.connection_exists("zone_II_inlet", "column_3"))

        self.assertTrue(flow_sheet.connection_exists("zone_III_inlet", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("zone_III_inlet", "column_1"))
        self.assertTrue(flow_sheet.connection_exists("zone_III_inlet", "column_2"))
        self.assertTrue(flow_sheet.connection_exists("zone_III_inlet", "column_3"))

        self.assertTrue(flow_sheet.connection_exists("zone_IV_inlet", "column_0"))
        self.assertTrue(flow_sheet.connection_exists("zone_IV_inlet", "column_1"))
        self.assertTrue(flow_sheet.connection_exists("zone_IV_inlet", "column_2"))
        self.assertTrue(flow_sheet.connection_exists("zone_IV_inlet", "column_3"))

        self.assertTrue(flow_sheet.connection_exists("column_0", "zone_I_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_0", "zone_II_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_0", "zone_III_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_0", "zone_IV_outlet"))

        self.assertTrue(flow_sheet.connection_exists("column_1", "zone_I_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_1", "zone_II_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_1", "zone_III_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_1", "zone_IV_outlet"))

        self.assertTrue(flow_sheet.connection_exists("column_2", "zone_I_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_2", "zone_II_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_2", "zone_III_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_2", "zone_IV_outlet"))

        self.assertTrue(flow_sheet.connection_exists("column_3", "zone_I_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_3", "zone_II_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_3", "zone_III_outlet"))
        self.assertTrue(flow_sheet.connection_exists("column_3", "zone_IV_outlet"))

    def test_column_position_indices(self):
        """Test column position indices."""
        builder = self.create_smb()

        # Initial state, position 0
        carousel_position = 0
        carousel_state = 0
        indices_expected = 0

        indices = builder.column_indices_at_state(carousel_position, carousel_state)
        self.assertEqual(indices_expected, indices)

        time = carousel_state * builder.switch_time
        indices = builder.column_indices_at_time(time, carousel_position)
        self.assertEqual(indices_expected, indices)

        # Initial state, position 1
        carousel_position = 1
        carousel_state = 0
        indices_expected = 1

        indices = builder.column_indices_at_state(carousel_position, carousel_state)
        self.assertEqual(indices_expected, indices)

        time = carousel_state * builder.switch_time
        indices = builder.column_indices_at_time(time, carousel_position)
        self.assertEqual(indices_expected, indices)

        # First state, position 0
        carousel_position = 0
        carousel_state = 1
        indices_expected = 1

        indices = builder.column_indices_at_state(carousel_position, carousel_state)
        self.assertEqual(indices_expected, indices)

        time = carousel_state * builder.switch_time
        indices = builder.column_indices_at_time(time, carousel_position)
        self.assertEqual(indices_expected, indices)

        # First state, position 1
        carousel_position = 1
        carousel_state = 1
        indices_expected = 2

        indices = builder.column_indices_at_state(carousel_position, carousel_state)
        self.assertEqual(indices_expected, indices)

        time = carousel_state * builder.switch_time
        indices = builder.column_indices_at_time(time, carousel_position)
        self.assertEqual(indices_expected, indices)

        # 4th state (back to initial state), position 0
        carousel_position = 0
        carousel_state = 4
        indices_expected = 0

        indices = builder.column_indices_at_state(carousel_position, carousel_state)
        self.assertEqual(indices_expected, indices)

        time = carousel_state * builder.switch_time
        indices = builder.column_indices_at_time(time, carousel_position)
        self.assertEqual(indices_expected, indices)

    def test_carousel_state(self):
        """Test carousel state."""

        builder = self.create_smb()

        # Initial state
        time = 0
        state_expected = 0

        state = builder.carousel_state(time)

        self.assertEqual(state_expected, state)

        # Position 0
        time = builder.switch_time / 2
        state_expected = 0

        state = builder.carousel_state(time)

        self.assertEqual(state_expected, state)

        # Position 1
        time = builder.switch_time
        state_expected = 1

        state = builder.carousel_state(time)

        self.assertEqual(state_expected, state)

        # Back to initial state; position 0
        time = 4 * builder.switch_time
        state_expected = 0

        state = builder.carousel_state(time)

        self.assertEqual(state_expected, state)

    def test_flow_rates(self):
        # Serial
        builder = self.create_serial()
        process = builder.build_process()

        serial_inlet = process.flow_rate_timelines["serial_inlet"]
        serial_outlet = process.flow_rate_timelines["serial_outlet"]
        column_0 = process.flow_rate_timelines["column_0"]
        column_1 = process.flow_rate_timelines["column_1"]

        flow_rate = serial_inlet.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = serial_outlet.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_0.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_1.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        # Parallel (flow is split between columns)
        builder = self.create_parallel()
        process = builder.build_process()

        parallel_inlet = process.flow_rate_timelines["parallel_inlet"]
        column_0 = process.flow_rate_timelines["column_0"]
        column_1 = process.flow_rate_timelines["column_1"]

        flow_rate = parallel_inlet.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = parallel_inlet.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_0.total_in[None].value(0)
        flow_rate_expected = 1e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_1.total_in[None].value(0)
        flow_rate_expected = 1e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        # Multi-Zone (Side streams between zones)
        builder = self.create_multi_zone()
        process = builder.build_process()

        serial_inlet = process.flow_rate_timelines["serial_inlet"]
        parallel_inlet = process.flow_rate_timelines["parallel_inlet"]
        column_0 = process.flow_rate_timelines["column_0"]
        column_2 = process.flow_rate_timelines["column_2"]

        flow_rate = serial_inlet.total_in[None].value(0)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = parallel_inlet.total_in[None].value(0)
        flow_rate_expected = 3e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        # Initial state
        t = 0
        flow_rate = column_0.total_in[None].value(t)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_2.total_in[None].value(t)
        flow_rate_expected = 1.5e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        # First position
        t = builder.switch_time
        flow_rate = column_0.total_in[None].value(t)
        flow_rate_expected = 1.5e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

        flow_rate = column_2.total_in[None].value(t)
        flow_rate_expected = 2e-7
        np.testing.assert_almost_equal(flow_rate, flow_rate_expected)

    def test_flow_direction(self):
        # Multi-Zone (Side streams between zones)
        builder = self.create_multi_zone()
        process = builder.build_process()

        # Initial state
        t = 0

        tl = process.parameter_timelines["flow_sheet.column_0.flow_direction"]
        flow_direction = tl.value(t)
        flow_direction_expected = 1
        np.testing.assert_almost_equal(flow_direction, flow_direction_expected)

        tl = process.parameter_timelines["flow_sheet.column_2.flow_direction"]
        flow_direction = tl.value(t)
        flow_direction_expected = -1
        np.testing.assert_almost_equal(flow_direction, flow_direction_expected)

        # First position
        t = builder.switch_time

        tl = process.parameter_timelines["flow_sheet.column_0.flow_direction"]
        flow_direction = tl.value(t)
        flow_direction_expected = -1
        np.testing.assert_almost_equal(flow_direction, flow_direction_expected)

        tl = process.parameter_timelines["flow_sheet.column_2.flow_direction"]
        flow_direction = tl.value(t)
        flow_direction_expected = 1
        np.testing.assert_almost_equal(flow_direction, flow_direction_expected)

    def test_simulation(self):
        builder = self.create_serial()
        process = builder.build_process()

        process_simulator = Cadet()
        simulation_results = process_simulator.simulate(process)

        self.assertEqual(simulation_results.exit_flag, 0)


if __name__ == "__main__":
    unittest.main()



from typing import Any

import numpy as np

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import Structure, UnsignedFloat, UnsignedInteger, Vector

__all__ = ["Fraction", "FractionPool"]


class Fraction(Structure):
    """
    A class representing a fraction of a mixture.

    A fraction is defined by its mass and volume, and can be used to calculate
    properties such as cumulative mass, purity, and concentration.

    Attributes
    ----------
    mass : np.ndarray
        Mass of each component in the fraction.
    volume : float
        Volume of the fraction.

    Properties
    ----------
    n_comp : int
        Number of components in the fraction.
    fraction_mass : np.ndarray
        Cumulative mass of all species in the fraction.
    purity : np.ndarray
        Purity of the fraction, with invalid values set to zero.
    concentration : np.ndarray
        Component concentrations of the fraction, with invalid values set to zero.

    See Also
    --------
    CADETProcess.fractionation.FractionPool
    CADETProcess.fractionation.Fractionator
    """

    mass = Vector()
    volume = UnsignedFloat()
    start = UnsignedFloat()
    end = UnsignedFloat()

    _parameters = ["mass", "volume", "start", "end"]

    @property
    def n_comp(self) -> int:
        """int: Number of components in the fraction."""
        return self.mass.size

    @property
    def fraction_mass(self) -> np.ndarray:
        """
        np.ndarray: Cumulative mass all species in the fraction.

        See Also
        --------
        mass
        purity
        concentration
        """
        return sum(self.mass)

    @property
    def purity(self) -> np.ndarray:
        """
        np.ndarray: Purity of the fraction.

        Invalid values are replaced by zero.

        See Also
        --------
        mass
        fraction_mass
        concentration
        """
        with np.errstate(divide="ignore", invalid="ignore"):
            purity = self.mass / self.fraction_mass

        return np.nan_to_num(purity)

    @property
    def concentration(self) -> np.ndarray:
        """
        np.ndarray: Component concentrations of the fraction.

        Invalid values are replaced by zero.

        See Also
        --------
        mass
        volume
        """
        with np.errstate(divide="ignore", invalid="ignore"):
            concentration = self.mass / self.volume

        return np.nan_to_num(concentration)

    def __repr__(self) -> str:
        """str: String representation of the fraction."""
        return f"{self.__class__.__name__}(mass={self.mass},volume={self.volume})"


class FractionPool(Structure):
    """
    Collection of pooled fractions.

    This class manages multiple fractions of a mixture, facilitating the
    calculation of cumulative properties of the pool, such as total volume,
    total mass, average purity, and average concentration.

    Attributes
    ----------
    n_comp : int
        The number of components each fraction in the pool should have.

    Properties
    ----------
    fractions : list
        List of fractions in the pool.
    n_fractions : int
        Number of fractions in the pool.
    volume : float
        Total volume of all fractions in the pool.
    mass : np.ndarray
        Cumulative mass of each component in the pool.
    pool_mass : float
        Total mass of all components in the pool.
    purity : np.ndarray
        Overall purity of the pool, with invalid values set to zero.
    concentration : np.ndarray
        Average concentration of the pool, with invalid values set to zero.

    See Also
    --------
    CADETProcess.fractionation.Fraction
    CADETProcess.fractionation.Fractionator
    """

    n_comp = UnsignedInteger()

    _parameters = ["n_comp"]

    def __init__(self, n_comp: int, *args: Any, **kwargs: Any) -> None:
        """
        Initialize a FractionPool instance.

        Parameters
        ----------
        n_comp : int
            The number of components each fraction in the pool should have.
        *args : Optional
            Optional Parameters for Structure class.
        **kwargs : Optional
            Additional Parameters for Structure class.
        """
        self._fractions = []
        self.n_comp = n_comp

        super().__init__(*args, **kwargs)

    def add_fraction(self, fraction: Fraction) -> None:
        """
        Add a fraction to the fraction pool.

        Parameters
        ----------
        fraction : Fraction
            The fraction to be added to the pool.

        Raises
        ------
        CADETProcessError
            If the fraction is not an instance of the Fraction class, or if
            the number of components in the fraction does not match the number
            of components in the pool.
        """
        if not isinstance(fraction, Fraction):
            raise CADETProcessError("Expected Fraction")

        if fraction.n_comp != self.n_comp:
            raise CADETProcessError("Number of components does not match.")

        self._fractions.append(fraction)

    @property
    def fractions(self) -> list[Fraction]:
        """list: List of fractions in the pool."""
        if len(self._fractions) == 0:
            return [Fraction(np.zeros((self.n_comp,)), 0)]
        return self._fractions

    @property
    def n_fractions(self) -> int:
        """int: Number of fractions in the pool."""
        return len(self._fractions)

    @property
    def volume(self) -> float:
        """float: Sum of all fraction volumes in the fraction pool."""
        return sum(frac.volume for frac in self.fractions)

    @property
    def mass(self) -> np.ndarray:
        """np.ndarray: Cumulative component mass in the fraction pool."""
        return np.sum([frac.mass for frac in self.fractions], axis=0)

    @property
    def pool_mass(self) -> float:
        """float: Sum of cumulative component mass in the fraction pool."""
        return sum(frac.fraction_mass for frac in self.fractions)

    @property
    def purity(self) -> np.ndarray:
        """
        Total purity of components in the fraction pool.

        Invalid values are replaced by zero.

        Returns
        -------
        purity : np.ndarray
            Purity of each component in the fraction pool.

        See Also
        --------
        mass
        pool_mass
        concentration
        """
        with np.errstate(divide="ignore", invalid="ignore"):
            purity = self.mass / self.pool_mass

        return np.nan_to_num(purity)

    @property
    def concentration(self) -> np.ndarray:
        """
        Total concentration of components in the fraction pool.

        Invalid values are replaced by zero.

        Returns
        -------
        concentration : np.ndarray
            Average concentration of the fraction pool.

        See Also
        --------
        mass
        volume
        """
        with np.errstate(divide="ignore", invalid="ignore"):
            concentration = self.mass / self.volume

        return np.nan_to_num(concentration)

    def __repr__(self) -> str:
        """str: String representation of the fraction pool."""
        return f"{self.__class__.__name__}(n_comp={self.n_comp})"
